# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from accelerator.colourwrapper import colour
from accelerator.compat import fmt_num, num_types, int_types
from accelerator.error import NoSuchWhateverError
from accelerator.extras import quote
from accelerator.shell.parser import ArgumentParser
from accelerator.shell.parser import name2ds
from accelerator import g

from collections import Counter
import math
import sys


# multiprocessing.Pool insists on pickling the function even when using
# forking (as we do). Thus we need this top level indirection function.
_indirected_func = None
def _indirection(sliceno):
	return _indirected_func(sliceno)


NaN = float('NaN')
def merge_nans(hist):
	# NaN != NaN, so they are not collected in a single key.
	# Unless they are the same object, like our global above.
	# dsutil also uses only a single (different) NaN object.
	# (But that does not survive pickling.)
	nan_count = 0
	for k, v in list(hist.items()):
		if isinstance(k, float) and math.isnan(k):
			nan_count += v
			del hist[k]
	if nan_count:
		hist[NaN] = nan_count
		# Now all NaNs are under a single key.
	return hist


def format_aligned(hist):
	def fmt_k(k):
		if isinstance(k, num_types):
			return fmt_num(k)
		else:
			return str(k).expandtabs()
	empty = [v == 0 for k, v in hist]
	hist = [(fmt_k(k), fmt_num(v)) for k, v in hist]
	klen = max(len(k) for k, v in hist)
	vlen = max(len(v) for k, v in hist)
	total_len = klen + vlen + 2
	hist = [(k, ' ' * (total_len - len(k) - len(v)), v) for k, v in hist]
	hist = [
		(colour(k, 'hist/empty') if e else k, s, colour(v, 'hist/empty') if e else v)
		for e, (k, s, v) in zip(empty, hist)
	]
	return hist, colour('%s', 'hist/range') + '%s' + colour('%s', 'hist/count')

def format_bars(hist):
	from accelerator.compat import terminal_size
	columns = terminal_size()[0]
	a_hist, a_fmt = format_aligned(hist)
	max_len = columns - len(''.join(a_hist[0])) - 3
	if max_len < 2:
		return a_hist, a_fmt
	if sys.stdout.encoding.lower().startswith('utf-8'):
		characters = ('', '\u258f', '\u258e', '\u258d', '\u258c', '\u258b', '\u258a', '\u2589', '\u2588')
	else:
		characters = ('', '', '', '', '', '#', '#', '#', '#')
	values = [v for k, v in hist]
	max_v = max(values)
	if max_v > max_len:
		def mkbar(v):
			size = v * max_len / max_v
			bonus = int(round((size - int(size)) * 8))
			return characters[-1] * int(size) + characters[bonus]
	else:
		mkbar = characters[-1].__mul__
	hist = [(a_fmt % a, mkbar(v)) for a, v in zip(a_hist, values)]
	return hist, '%s  ' + colour('%s', 'hist/bar')

def _escape(hist, char):
	def q(s):
		if char in s or s.startswith('"'):
			return '"' + s.replace('"', '""') + '"'
		else:
			return s
	return [(q(str(k)), v) for k, v in hist]

def format_csv(hist):
	return _escape(hist, ','), '%s,%d'

def format_ssv(hist):
	return _escape(hist, ';'), '%s;%d'

def format_tsv(hist):
	return _escape(hist, '\t'), '%s\t%d'

formatters = {
	'aligned': format_aligned,
	'bars'   : format_bars,
	'csv'    : format_csv,
	'ssv'    : format_ssv,
	'tsv'    : format_tsv,
}


def main(argv, cfg):
	parser = ArgumentParser(prog=argv.pop(0), description='''show a histogram of column(s) from a dataset.''')
	parser.add_argument('-c', '--chain',     action='store_true', negation='dont', help="follow dataset chain", )
	parser.add_argument(      '--chain-length', '--cl',         metavar='LENGTH',  help="follow chain at most this many datasets", type=int)
	parser.add_argument(      '--stop-ds',   metavar='DATASET', help="follow chain at most to this dataset")
	parser.add_argument('-f', '--format', choices=sorted(formatters), help="output format, " + ' / '.join(sorted(formatters)), metavar='FORMAT', )
	parser.add_argument('-t', '--toplist',   action='store_true', negation='not',  help="don't bin values, show most common")
	parser.add_argument('-m', '--max-count', metavar='NUM',     help="show at most this many values or bins (default ~ 20)", type=int)
	parser.add_argument('-s', '--slice',     action='append',   help="this slice only, can be specified multiple times", type=int)
	parser.add_argument('dataset', help='can be specified in the same ways as for "ax ds"')
	parser.add_argument('column', nargs='*', help='you can specify multiple columns')
	args = parser.parse_intermixed_args(argv)

	if args.max_count and args.max_count < 0:
		args.max_count = 0

	try:
		ds = name2ds(cfg, args.dataset)
	except NoSuchWhateverError as e:
		print(e, file=sys.stderr)
		return 1

	chain = ds.chain(args.chain_length if args.chain else 1, stop_ds=args.stop_ds)
	if not chain:
		return

	if not args.column:
		# Can we guess?
		columns_in_all = set(chain[0].columns)
		for ds in chain:
			columns_in_all.intersection_update(ds.columns)
		if len(columns_in_all) == 1:
			args.column = list(columns_in_all)
		else:
			print('Specify at least one column, choose from:', file=sys.stderr)
			for colname in sorted(columns_in_all):
				print('    ' + quote(colname), file=sys.stderr)
			return 1

	ok = True
	for ds in chain:
		for col in args.column:
			if col not in ds.columns:
				print("Dataset %s does not have column %s." % (ds.quoted, col,), file=sys.stderr)
				ok = False
	if not ok:
		return 1

	varying_bucket_size = False

	columns = args.column[0] if len(args.column) == 1 else args.column
	useful_slices = [sliceno for sliceno in range(g.slices) if chain.lines(sliceno) > 0]
	if args.slice:
		useful_slices = list(set(useful_slices) & set(args.slice))
	if not useful_slices:
		return

	def mp_run(part_func, collect_func):
		if len(useful_slices) == 1:
			return collect_func(None, part_func(useful_slices[0]))
		from multiprocessing import Pool
		global _indirected_func
		_indirected_func = part_func
		pool = Pool(len(useful_slices))
		try:
			collected = None
			for part in pool.imap_unordered(_indirection, useful_slices, 1):
				collected = collect_func(collected, part)
			return collected
		finally:
			pool.close()

	def count_items(sliceno):
		return Counter(chain.iterate(sliceno, columns))
	hist_func = count_items # default, probably overridden if not args.toplist

	if len(args.column) > 1 or chain.min(columns) is None:
		args.toplist = True # Can't do anything else.

	if not args.toplist:
		low, high = chain.min(columns), chain.max(columns)
		if high == low: # just one value, so let's not try to bin it
			args.toplist = True

	if not args.toplist:
		inf = float('inf')
		if low in (inf, -inf) or high in (inf, -inf):
			print("Can't bin to infinity.", file=sys.stderr)
			return 1
		if args.max_count != 0: # otherwise all values are their own bin
			non_ints = chain.filter(lambda ds: not ds.columns[columns].type.startswith('int'))
			is_ints = not non_ints
			if non_ints and all(ds.columns[columns].type == 'number' for ds in non_ints):
				class Failure(Exception):
					pass
				def check_int(sliceno):
					return all(isinstance(v, int_types) for v in non_ints.iterate(sliceno, columns))
				def collect_intness(dummy, good):
					if not good:
						raise Failure() # abort the other slices, for speed
				try:
					mp_run(check_int, collect_intness)
					is_ints = True
				except Failure:
					is_ints = False
			if args.max_count is None:
				range_len = high - low + is_ints
				cands = (20, 21, 19, 22, 18, 23, 17, 24, 16, 25, 15, 14, 13,)
				if is_ints:
					# try to find something that gives equal sized buckets.
					args.max_count = 20 # fallback
					for cand in cands:
						if range_len / cand == range_len // cand:
							args.max_count = cand
							break
				else:
					# find the shortest string representation of the step,
					# as that is likely to look better in the labels.
					best = (1e1000, 20)
					for cand in cands:
						score = len(str(range_len / cand))
						if score < best[0]:
							best = (score, cand)
					args.max_count = best[1]
			if is_ints and high - low < args.max_count:
				args.max_count = high - low + 1
			step = (high - low) / args.max_count
			if is_ints and (high - low + 1) / args.max_count != math.ceil(step):
				varying_bucket_size = True
			def bin_items(sliceno):
				return merge_nans(Counter((v - low) // step for v in chain.iterate(sliceno, columns)))
			hist_func = bin_items
			if is_ints:
				def name(ix):
					a = int(math.ceil(step * ix + low))
					b = step * (ix + 1) + low
					if b == math.floor(b) and ix < args.max_count - 1:
						b = int(b) - 1
					else:
						b = int(b)
					if a == b:
						return fmt_num(a)
					else:
						return '%s - %s' % (fmt_num(a), fmt_num(b))
			else:
				def name(ix):
					a = step * ix + low
					b = step * (ix + 1) + low
					return '%s - %s' % (fmt_num(a), fmt_num(b))
			bin_names = [name(ix) for ix in range(args.max_count)]

	def collect_hist(hist, part):
		if not hist:
			return part
		hist.update(part)
		return hist

	hist = merge_nans(mp_run(hist_func, collect_hist))

	if args.format:
		formatter = formatters[args.format]
	elif sys.stdout.isatty():
		formatter = format_bars
	else:
		formatter = format_tsv

	if args.toplist:
		total_found = len(hist)
		hist = hist.most_common(args.max_count or None)
		hist, fmt = formatter(hist)
	else:
		total_found = 0 # so we don't print about it later
		if args.max_count:
			if NaN in hist:
				print(colour('WARNING: Ignored %d NaN values.' % (hist[NaN],), 'hist/warning'), file=sys.stderr)
			hist[args.max_count - 1] += hist[args.max_count] # top value should not be in a separate bin
			hist, fmt = formatter([(name, hist[ix]) for ix, name in enumerate(bin_names)])
		else:
			hist, fmt = formatter([(k, hist[k]) for k in sorted(hist)])

	for item in hist:
		print(fmt % item)

	if varying_bucket_size:
		print(colour('WARNING: varying bucket size.', 'hist/warning'), file=sys.stderr)

	if total_found > len(hist):
		print(colour(fmt_num(total_found) + ' total found', 'hist/warning'), file=sys.stderr)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import unicode_literals

description = r'''
Test the +None types in dataset_type.

Types two datasets, one with
	['real value', '', ' ']
in each row and one with
	[None, 'real value', '', ' '].

These are both typed with and without +None on the types, with defaults
set to the first value.

Most types reject empty strings (including space), so the results should
mostly be
	([None] +) [value, None, value]
and
	([None] +) [value, value, value]

The resulting columns should have None support if typed as +None or
if the source had None support, but not otherwise.
'''

from accelerator import subjobs
from datetime import date, time, datetime

def nonename(name):
	a = name.split(':')
	a[0] += '+None'
	return ':'.join(a)

def synthesis(job):
	# {type: (string, typed value, overrides for ['', ' '])}
	# Not testing every type, but all implementations.
	types2test = {
		'ascii:encode'     : ('\u00fe', '\\303\\276', [ '',  ' '],),
		'asciistrip'       : (    '  ',           '', [ '',   ''],),
		'bytes'            : ('\u00fe',  b'\xc3\xbe', [b'', b' '],),
		'unicode:utf-8'    : ('\u00fe',     '\u00fe', [ '',  ' '],),
		'complex64'        : ('1-0.2j',     (1-0.2j), None,),
		'float32'          : (    '42',         42.0, None,),
		'float32i'         : (  '4two',          4.0, [0, 0],),
		'float64'          : (   '4.2',         4.20, None,),
		'float64i'         : (  '.4.2',          0.4, [0, 0],),
		'floatbool'        : (     '9',         True, None,),
		'floatbooli'       : (     'y',        False, [False, False],),
		'int32_10'         : (    '18',           18, None,),
		'int64_16i'        : (  'bark',         0xba, [0, 0],),
		'int64_8'          : (  '1000',          512, None,),
		'json'             : ('[1, 2]',       [1, 2], None,),
		'number'           : (   '1.8',          1.8, None,),
		'number:int'       : (     '7',            7, None,),
		'strbool'          : (    'no',        False, [False, True],), # everything unrecognised is True, so ' ' is True
		'datetime:%Y'      : (  '2008', datetime(2008, 1, 1), None,),
		'datetimei:%Y'     : (  '200x', datetime( 200, 1, 1), None,),
		'date:%Y'          : (  '1997', date(1997, 1, 1)    , None,),
		'datei:%Y'         : (  '199x', date( 199, 1, 1)    , None,),
		'time:%H:%M'       : ( '18:27', time(18, 27)        , None,),
		'timei:%H:%M'      : ( '18:2x', time(18,  2)        , None,),
	}
	defaults = {k: v[0] for k, v in types2test.items()}
	for ds_with_None in (False, True):
		dw = job.datasetwriter(
			name='with_None' if ds_with_None else 'without_None',
			allow_missing_slices=True,
		)
		for name in types2test:
			dw.add(name, 'unicode', none_support=ds_with_None)
		dw.set_slice(0)
		if ds_with_None:
			dw.write_dict({k: None for k in types2test})
		dw.write_dict({k: v[0] for k, v in types2test.items()})
		dw.write_dict({k: '' for k in types2test})
		dw.write_dict({k: ' ' for k in types2test})
		ds = dw.finish()
		for type_with_None in (False, True):
			column2type = {k: [str, nonename][type_with_None](k) for k in types2test}
			typed = subjobs.build('dataset_type', source=ds, column2type=column2type, defaults=defaults).dataset()
			for name, (_, value, values4empty) in types2test.items():
				assert typed.columns[name].none_support == ds_with_None or type_with_None
				if values4empty:
					want = [value] + values4empty
				else:
					want = [value, value, value]
				if type_with_None:
					want[1] = None
				if ds_with_None:
					want.insert(0, None)
				got = list(typed.iterate(0, name))
				assert want == got, 'Column %r in %s has %r, should have %r' % (name, typed.quoted, got, want,)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2020-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from traceback import print_exc
from datetime import datetime
import errno
import json
from argparse import RawTextHelpFormatter
import os
import sys

from accelerator.build import fmttime
from accelerator.colourwrapper import colour
from accelerator.error import NoSuchJobError
from accelerator.setupfile import encode_setup
from accelerator.compat import FileNotFoundError, url_quote, urlencode
from accelerator.unixhttp import call
from .parser import name2job, ArgumentParser

def show(url, job, verbose, show_output):
	setup = job.json_load('setup.json')
	is_build = setup.get('is_build', False)
	if verbose:
		print(colour(job.path, 'job/header'))
		print(colour('=' * len(job.path), 'job/header'))
		setup.pop('_typing', None)
		setup.starttime = str(datetime.fromtimestamp(setup.starttime))
		if 'endtime' in setup:
			setup.endtime = str(datetime.fromtimestamp(setup.endtime))
		print(encode_setup(setup, as_str=True))
	else:
		starttime = datetime.fromtimestamp(setup.starttime).replace(microsecond=0)
		hdr = '%s (%s) at %s' % (job, job.method, starttime,)
		if 'exectime' in setup:
			hdr = '%s in %s' % (hdr, fmttime(setup.exectime.total),)
		print(colour(hdr, 'job/header'))
		if job.is_build:
			print(colour('  build job', 'job/highlight'))
		if job.parent:
			built_from = "  built from %s (%s)" % (job.parent, job.parent.method,)
			if job.build_job and job.parent != job.build_job:
				built_from = "%s, build job %s (%s)" % (built_from, job.build_job, job.build_job.method,)
			print(colour(built_from, 'job/highlight'))
		things = []
		def opt_thing(name):
			value = setup[name]
			if value:
				if isinstance(value, dict) and len(value) > 1:
					value = json.dumps(value, indent=4, sort_keys=True).replace('\n', '\n    ')
				else:
					value = json.dumps(value)
				things.append((name, value,))
		opt_thing('caption')
		opt_thing('options')
		opt_thing('datasets')
		opt_thing('jobs')
		for k, v in things:
			print('    %s: %s' % (k, v,))
	def list_of_things(name, things):
		total = len(things)
		if total > 5 and not verbose:
			things = things[:3]
		print()
		print(colour('%s:' % (name,), 'job/header'))
		for thing in things:
			print('   ', thing)
		if total > len(things):
			print('    ... and %d more' % (total - len(things),))
	if job.datasets:
		list_of_things('datasets', [ds.quoted for ds in job.datasets])
	try:
		post = job.json_load('post.json')
	except FileNotFoundError:
		print(colour('WARNING: Job did not finish', 'job/warning'))
		post = None
	if post and post.subjobs:
		postdata = urlencode({'jobs': '\0'.join(post.subjobs)}).encode('utf-8')
		subjobs = call(url + '/jobs_are_current', data=postdata)
		fmted_subjobs = []
		for sj, is_current in sorted(subjobs.items()):
			if not is_current:
				sj += ' ' + colour('(not current)', 'job/info')
			fmted_subjobs.append(sj)
		list_of_things('subjobs', fmted_subjobs)
	if post and post.files:
		files = sorted(post.files)
		if verbose:
			files = [job.filename(fn) for fn in files]
		list_of_things('files', files)
	if post and not is_build and not call(url + '/job_is_current/' + url_quote(job)):
		print(colour('Job is not current', 'job/info'))
	print()
	out = job.output('parts')
	if show_output:
		if out:
			if not verbose: # verbose prints section headers in show_output_d()
				print(colour('output:', 'job/header'))
			show_output_d(out, verbose)
		else:
			print(job, 'produced no output')
			print()
	elif out:
		print('%s produced %d bytes of output, use --output/-o to see it' % (job, sum(len(v) for v in out.values()),))
		print()

def show_source(job, pattern='*'):
	import tarfile
	from fnmatch import fnmatch
	with tarfile.open(job.filename('method.tar.gz'), 'r:gz') as tar:
		all_members = [info for info in tar.getmembers() if info.isfile()]
		members = [info for info in all_members if fnmatch(info.path, pattern)]
		if not members:
			if pattern:
				print(colour('No sources matching %r in %s.' % (pattern, job,), 'job/warning'), file=sys.stderr)
				fh = sys.stderr
				res = 1
			else:
				fh = sys.stdout
				res = 0
			print('Available sources:', file=fh)
			for info in all_members:
				print('    ' + info.path, file=fh)
			return res
		for ix, info in enumerate(members, 1):
			if len(members) > 1:
				print(colour(info.path, 'job/header'))
				print(colour('=' * len(info.path), 'job/header'))
			data = tar.extractfile(info).read()
			has_nl = data.endswith(b'\n')
			while data:
				data = data[os.write(1, data):]
			if not has_nl:
				os.write(1, b'\n')
			if ix < len(members):
				os.write(1, b'\n')
	return 0

def show_file(job, pattern):
	try:
		files = job.files(pattern)
	except FileNotFoundError:
		# Probably the job didn't finish, let's not explode
		files = []
	if not files and pattern and os.path.exists(job.filename(pattern)):
		# special case so you can name unregistered files
		files = [pattern]
	if not files:
		if pattern:
			fh = sys.stderr
			print(colour('No files matching %r in %s.' % (pattern, job,), 'job/warning'), file=fh)
			res = 1
		else:
			fh = sys.stdout
			res = 0
		try:
			files = job.files()
		except FileNotFoundError:
			print(colour('Job did not finish - unable to list files', 'job/warning'), file=sys.stderr)
			return 1
		print('Available files:', file=fh)
		for fn in files:
			print('    ' + fn, file=fh)
		return res
	for ix, fn in enumerate(files, 1):
		if len(files) > 1:
			print(colour(fn, 'job/header'))
			print(colour('=' * len(fn), 'job/header'))
		with job.open(fn, 'rb') as fh:
			data = fh.read()
		has_nl = data.endswith(b'\n')
		while data:
			data = data[os.write(1, data):]
		if not has_nl:
			os.write(1, b'\n')
		if ix < len(files):
			os.write(1, b'\n')
	return 0

def show_output_d(d, verbose):
	first = True
	for k, out in d.items():
		if out:
			if verbose:
				if first:
					first = False
				else:
					print()
				if isinstance(k, int):
					k = 'analysis(%d)' % (k,)
				print(colour(k, 'job/header'))
				print(colour('=' * len(k), 'job/header'))
			print(out, end='' if out.endswith('\n') else '\n')

def main(argv, cfg):
	descr = 'show setup.json, dataset list, etc for jobs'
	parser = ArgumentParser(
		prog=argv.pop(0),
		description=descr,
		formatter_class=RawTextHelpFormatter,
	)
	parser.add_argument('-v', '--verbose', action='store_true', negation='not', help='more output (e.g the whole setup.json)')
	group = parser.add_mutually_exclusive_group()
	group.add_argument('-o', '--output', action='store_true', help='show job output')
	group.add_argument('-O', '--just-output', action='store_true', help='show only job output')
	group.add_argument('-P', '--just-path', action='store_true', help='show only job path')
	group.add_argument('-s', '--source', action='store_true', help='show source used to produce job')
	group.add_argument('-S', '--source-file', metavar='PATTERN', nargs='?', const='', help='show specified file(s) from source used to produce job')
	group.add_argument('-f', '--file', metavar='PATTERN', nargs='?', const='', help='show specified file(s) produced by job')
	parser.add_argument(
		'jobid',
		nargs='+', metavar='jobid/jobspec',
		help='jobid is just a jobid.\n' +
		     'you can also use path, method or :urdlist:[entry].\n' +
		     'path is to a jobdir (with setup.json in it).\n' +
		     'method is the latest job with that method.\n' +
		     ':urdlist:[entry] looks up jobs in urd. details are in the\n' +
		     'urd help, except here entry defaults to -1 and you can\'t\n' +
		     'list things (no .../ or .../since/x).\n' +
		     'you can use spec~ or spec~N to go back N jobs\n' +
		     'with that method or spec^ or spec^N to follow .previous\n' +
		     'use spec! to only consider current jobs.\n' +
		     'you can also do things like spec.source.'
	)
	args = parser.parse_intermixed_args(argv)
	res = 0
	for path in args.jobid:
		try:
			job = name2job(cfg, path)
			if args.just_output:
				show_output_d(job.output('parts'), args.verbose)
			elif args.just_path:
				print(job.path)
			elif args.source:
				res |= show_source(job)
			elif args.source_file is not None:
				res |= show_source(job, args.source_file)
			elif args.file is not None:
				res |= show_file(job, args.file)
			else:
				show(cfg.url, job, args.verbose, args.output)
		except NoSuchJobError as e:
			print(e)
			res = 1
		except Exception as e:
			if isinstance(e, OSError) and e.errno == errno.EPIPE:
				raise
			print_exc(file=sys.stderr)
			print("Failed to show %r" % (path,), file=sys.stderr)
			res = 1
	return res



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import unicode_literals

description = r'''
Verify that csvimport puts lines in the expected slice.

Both the actual data and the skipped lines should roundrobin independantly
starting from slice 0.

Does not test allow_bad, as that does not currenctly have a useful rule
for how data is sliced.
'''

from accelerator import subjobs
from itertools import cycle
from random import randint

def check(job, name, labels_txt=[], initial_skipped=[], sometimes='', comments=False, **kw):
	want_linenos = [[] for _ in range(job.params.slices)]
	lineno = [1]
	want_comments = [[] for _ in range(job.params.slices)]
	comments_cycle = cycle(want_comments)
	def add_comment(line):
		next(comments_cycle).append((lineno[0], line.encode('ascii')))
		fh.write(line + '\n')
		lineno[0] += 1
	with job.open(name, 'wt') as fh:
		for line in initial_skipped:
			add_comment(line)
		for line in labels_txt:
			fh.write(line + '\n')
			lineno[0] += 1
		for ix in range(10):
			for sliceno in range(job.params.slices):
				if sometimes and randint(0, 9) == 3:
					add_comment(sometimes % (ix, sliceno,))
				if comments and sliceno % comments == 0:
					add_comment('# line %d, before %d,%d' % (lineno[0], ix, sliceno,))
				fh.write('%d,%d\n' % (ix, sliceno,))
				want_linenos[sliceno].append(lineno[0])
				lineno[0] += 1
	job = subjobs.build(
		'csvimport',
		filename=job.filename(name),
		lineno_label='lineno',
		comment='#' if comments or sometimes else '',
		skip_lines=len(initial_skipped),
		**kw
	)

	ds = job.dataset()
	for sliceno in range(job.params.slices):
		got = [int(x) for x in ds.iterate(sliceno, 'sliceno')]
		assert got == [sliceno] * 10, "Slice %d has wrong slices in %s: %r" % (sliceno, ds.quoted, got,)
		got = [int(x) for x in ds.iterate(sliceno, 'ix')]
		assert got == list(range(10)), "Slice %d has wrong ixes in %s: %r" % (sliceno, ds.quoted, got,)
		got = list(ds.iterate(sliceno, 'lineno'))
		assert got == want_linenos[sliceno], "Slice %d has wrong lines in %s:\n    wanted %r\n    got    %r" % (sliceno, ds.quoted, want_linenos[sliceno], got,)
	if comments or sometimes or initial_skipped:
		ds = job.dataset('skipped')
		for sliceno in range(job.params.slices):
			got = list(ds.iterate(sliceno, ('lineno', 'data')))
			assert got == want_comments[sliceno], "Slice %d has wrong skipped lines in %s:\n    wanted %r\n    got    %r" % (sliceno, ds.quoted, want_comments[sliceno], got,)


def synthesis(job):
	check(job, 'simple.txt', labels=['ix', 'sliceno'])
	check(job, 'labeled.txt', labels_txt=['ix,sliceno'])
	check(job, 'comments.txt', labels_txt=['ix,sliceno'], comments=True)
	check(job, 'some comments.txt', labels_txt=['ix,sliceno'], comments=3)
	check(job, 'random comments.txt', labels_txt=['ix,sliceno'], sometimes='# a comment before %d,%d')
	check(job, 'two label lines.txt', labels_txt=['i,slice', 'x,no'], label_lines=2, rename={'i x': 'ix', 'slice no': 'sliceno'})
	check(job, 'two label lines after comments.txt', labels_txt=['i,slice', 'x,no'], label_lines=2, rename={'i x': 'ix', 'slice no': 'sliceno'}, initial_skipped=['this is some text', 'this is more text'], comments=2)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = """
Test job.save() and job.json_save() with background=True
using silly objects that sleep to make saving slow without a lot of IO.

Quite similar to test_job_save other than that.

Tests that they can be loaded in the next execution steps, and that the
save call returned faster than the sleep time.

The current implementation waits between the execution steps, but the API
does not promise that. This test is intended to remain valid if that is
fixed, and tests that the returned objects handle having .wait() called in
a different step (and also that .load() works without .wait()).

Should not take much more than 3 * sleeptime to run.
"""

options = {'sleeptime': 0.5}

from accelerator.compat import monotonic
from accelerator.compat import unicode
import time

# This pickles as a string, but slowly.
# It's a hack to make pickle encoding slow.
class SlowToPickle(object):
	__slots__ = ('text', 'sleeptime',)

	def __init__(self, text, sleeptime):
		self.text = text
		self.sleeptime = sleeptime

	def __reduce__(self):
		time.sleep(self.sleeptime)
		return unicode, (unicode(self.text),)

# This is a True boolean that takes a long time to evaluate.
# It's a hack to make json encoding slow.
class SlowTrue(object):
	__slots__ = ('sleeptime',)

	def __init__(self, sleeptime):
		self.sleeptime = sleeptime

	# python2 version
	def __nonzero__(self):
		time.sleep(self.sleeptime)
		return True

	# python3 version
	def __bool__(self):
		time.sleep(self.sleeptime)
		return True

def save(job, name, sliceno):
	before = monotonic()
	p = job.save('contents of %s %s' % (name, sliceno,), name + '.pickle', sliceno=sliceno)
	j = job.json_save({name: sliceno}, name + '.json', sliceno=sliceno)
	name = 'background ' + name
	bp = job.save(SlowToPickle('contents of %s %s' % (name, sliceno,), options.sleeptime), name + '.pickle', sliceno=sliceno, background=True)
	bj = job.json_save({name: sliceno}, name + '.json', sliceno=sliceno, sort_keys=SlowTrue(options.sleeptime), background=True)
	save_time = monotonic() - before
	max_time = options.sleeptime * 2 # two slow files
	assert save_time < max_time, "Saving took %s seconds, should have been less than %s" % (save_time, max_time,)
	return p, j, bp, bj

def check(job, name, sliceno, p, j, bp, bj, do_background=True, do_wait=False):
	todo = [(name, p, j)]
	if do_background:
		todo.append(('background ' + name, bp, bj))
	for name, p, j in todo:
		if do_wait:
			# Do explicit waiting sometimes
			p.wait()
			j.wait()
		assert p.load() == 'contents of %s %s' % (name, sliceno,)
		assert j.load() == {name: sliceno}
		for obj, filename in [(p, name + '.pickle'), (j, name + '.json')]:
			path = job.filename(filename, sliceno=sliceno)
			assert obj.path == path
			assert obj.filename == path.split('/')[-1]
			assert obj.jobwithfile().filename(sliceno) == path

def prepare(job):
	res = save(job, 'prepare', None)
	check(job, 'prepare', None, *res, do_background=False)

	# Test that the slowdowns work. (While the stuff we saved above is waiting)
	# Use sleeptime / 2.5 so this takes no extra time.
	checktime = options.sleeptime / 2.5
	before = monotonic()
	p = job.save(SlowToPickle('', checktime), 'test.pickle', background=True)
	p.wait()
	pickle_time = monotonic() - before
	assert pickle_time > checktime, "Saving a slow pickle took %s seconds, should have taken more than %s" % (pickle_time, checktime,)
	before = monotonic()
	j = job.json_save({}, 'test.json', sort_keys=SlowTrue(checktime), background=True)
	j.wait()
	json_time = monotonic() - before
	assert json_time > checktime, "Saving a slow json took %s seconds, should have taken more than %s" % (json_time, checktime,)

	return res

def analysis(sliceno, job, prepare_res):
	check(job, 'prepare', None, *prepare_res, do_wait=(sliceno % 2 == 0))
	res = save(job, 'analysis', sliceno)
	check(job, 'analysis', sliceno, *res, do_background=False)
	return res

def synthesis(job, prepare_res, analysis_res):
	check(job, 'prepare', None, *prepare_res, do_wait=True)
	for sliceno, res in enumerate(analysis_res):
		check(job, 'analysis', sliceno, *res, do_wait=(sliceno % 3 == 0))
	res = save(job, 'synthesis', None)
	check(job, 'synthesis', None, *res, do_wait=True)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2024 Carl Drougge                       #
# Modifications copyright (c) 2019-2021 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division, print_function

import sys
import locale
from math import ceil

from accelerator.compat import fmt_num, num_types
from accelerator.compat import terminal_size
from .parser import name2ds, name2job, ArgumentParser
from accelerator.colourwrapper import colour
from accelerator.error import NoSuchWhateverError
from accelerator.extras import quote

COLUMNS, LINES = terminal_size()


def colwidth(rows):
	# find max string len per column
	return [max(len(s) for s in col) for col in zip(*rows)]

def printcolwise(data, template, printfunc, minrows=8, indent=4):
	if not data:
		return
	cols = (COLUMNS - indent) // (len(template.format(*printfunc(data[0]))) + 2)
	n = int(ceil(len(data) / cols))
	n = max(n, minrows)
	for r in range(n):
		v = data[r::n]
		if v:
			print(' ' * indent + '  '.join(template.format(*printfunc(x)) for x in v))

def original_location(ds, colname):
	from accelerator.dataset import Dataset
	from accelerator.job import Job
	col = ds.columns[colname]
	if not col.location: # if it doesn't exist (0 line ds) we claim it comes from this ds
		return True, ds, colname
	parts = col.location.split('/')
	job = Job(parts[0])
	if job.version >= 4: # with /DS/, .[pm] and encoding of problem characters
		import re
		dsname = re.sub(r'\\x[0-9a-f]{2}', lambda m: chr(int(m.group()[2:], 16)), parts[2][:-2]).replace('\\\\', '\\')
	else:
		dsname = parts[1]
	src_ds = Dataset(job, dsname)
	for colname, cand_col in src_ds.columns.items():
		if cand_col.location == col.location and cand_col.offsets == col.offsets:
			return (src_ds == ds), src_ds, colname

def format_location(loc):
	if not loc:
		return '???' # should never happen
	is_local, ds, colname = loc
	if is_local:
		return 'local'
	return '%s in %s' % (quote(colname), ds.quoted,)

def typed_from(ds, loc):
	if not loc:
		return
	_, ds, colname = loc
	if ds.job.method != 'dataset_type':
		return
	src_ds = ds.job.params.datasets.source
	colname = unrename_column(ds.job, src_ds, colname)
	res = 'typed from ' + format_location((False, src_ds, colname))
	orig_loc = original_location(src_ds, colname)
	if orig_loc and not orig_loc[0]:
		return '%s, originally %s' % (res, format_location(orig_loc))
	else:
		return res

def unrename_column(type_job, ds, colname):
	rename = type_job.params.options.rename
	rev_rename = {v: k for k, v in rename.items() if k in ds.columns}
	return rev_rename.get(colname, colname)


def main(argv, cfg):
	usage = "%(prog)s [options] ds [ds [...]]"
	parser = ArgumentParser(prog=argv.pop(0), usage=usage)
	parser.add_argument('-c', '--chain',            action='store_true', negation='no',   help='list all datasets in a chain')
	parser.add_argument('-C', '--non-empty-chain',  action='store_true', negation='no',   help='list all non-empty datasets in a chain')
	parser.add_argument('-l', '--list',             action='store_true', negation='dont', help='list all datasets in a job with number of rows')
	parser.add_argument('-L', '--chainedlist',      action='store_true', negation='no',   help='list all datasets in a job with number of chained rows')
	parser.add_argument('-m', '--suppress-minmax',  action='store_true', negation='dont', help='do not print min/max column values')
	parser.add_argument('-n', '--suppress-columns', action='store_true', negation='dont', help='do not print columns')
	parser.add_argument('-q', '--suppress-errors',  action='store_true', negation='dont', help='silently ignores bad input datasets/jobids')
	parser.add_argument('-s', '--slices',           action='store_true', negation='no',   help='list relative number of lines per slice in sorted order')
	parser.add_argument('-S', '--chainedslices',    action='store_true', negation='no',   help='same as -s but for full chain')
	parser.add_argument('-w', '--location',         action='store_true', negation='no',   help='show original location of each column')
	parser.add_argument("dataset", nargs='+', help='the job part of the dataset name can be specified in the same ways as for "ax job". you can use ds~ or ds~N to follow the chain N steps backwards, or ^ to follow .parent. this requires specifying the ds-name, so wd-1~ will not do this, but wd-1/default~ will.')
	args = parser.parse_intermixed_args(argv)
	args.chain = args.chain or args.non_empty_chain

	def finish(badinput):
		if badinput and not args.suppress_errors:
			print('Error, failed to resolve datasets:', file=sys.stderr)
			for n, e in badinput:
				print('    %r: %s' % (n, e,), file=sys.stderr)
			exit(1)
		exit()

	badinput = []

	if args.list or args.chainedlist:
		for n in args.dataset:
			try:
				try:
					dsvec = name2ds(cfg, n).job.datasets
				except NoSuchWhateverError:
					dsvec = name2job(cfg, n).datasets
			except Exception as e:
				badinput.append((n, e))
				dsvec = None
			if dsvec:
				print('%s' % (dsvec[0].job,))
				v = []
				for ds in dsvec:
					if args.chainedlist:
						lines = sum(sum(x.lines) for x in ds.chain())
					else:
						lines = sum(ds.lines)
					v.append((ds.quoted, fmt_num(lines)))
				len_n, len_l = colwidth(v)
				template = "{0:%d}  ({1:>%d})" % (len_n, len_l)
				for name, numlines in sorted(v):
					print('    ' + template.format(name, numlines))
		finish(badinput)

	for n in args.dataset:
		try:
			ds = name2ds(cfg, n)
		except NoSuchWhateverError as e:
			badinput.append((n, e))
			continue

		print(ds.quoted)
		if ds.parent:
			if isinstance(ds.parent, tuple):
				print("    Parents:")
				max_n = max(len(x.quoted) for x in ds.parent)
				template = "{1:%d}" % (max_n,)
				data = tuple((None, x.quoted) for ix, x in enumerate(ds.parent))
				data = sorted(data, key = lambda x: x[1])
				printcolwise(data, template, lambda x: x, minrows=8, indent=8)
			else:
				print("    Parent:", ds.parent.quoted)
		print("    Method:", quote(ds.job.method))
		if ds.filename:
			print("    Filename:", quote(ds.filename))
		if ds.previous:
			print("    Previous:", ds.previous.quoted)
		if ds.hashlabel is not None:
			print("    Hashlabel:", quote(ds.hashlabel))

		if not args.suppress_columns:
			print("    Columns:")
			name2typ = {n: c.type + '+None' if c.none_support else c.type for n, c in ds.columns.items()}
			len_n, len_t = colwidth((quote(n), name2typ[n]) for n, c in ds.columns.items())
			if args.location:
				locations = {n: original_location(ds, n) for n in ds.columns}
				len_l = max(len(format_location(locations[n])) for n in ds.columns)
				len_c = max(len(c.compression) for c in ds.columns.values())
				template = '        {2} {0:%d}  {1:%d}  {4:%d}  {5:%d}  {3}' % (len_n, len_t, len_l, len_c,)
			else:
				template = '        {2} {0:%d}  {1:%d}  {3}' % (len_n, len_t,)
				locations = {}
			chain = ds.chain(-1 if args.chainedslices or args.chain else 1)

			def fmt_minmax(val):
				if isinstance(val, num_types):
					return fmt_num(val)
				else:
					return str(val)
			minlen = max(len(fmt_minmax(chain.min(n))) for n in ds.columns)
			maxlen = max(len(fmt_minmax(chain.max(n))) for n in ds.columns)
			minmax_template = '[%%%ds, %%%ds]' % (min(18, minlen), min(18, maxlen),)
			def prettyminmax(n):
				minval, maxval = chain.min(n), chain.max(n)
				if args.suppress_minmax or minval is None:
					return ''
				return minmax_template % (fmt_minmax(minval), fmt_minmax(maxval),)

			for n, c in sorted(ds.columns.items()):
				hashdot = colour("*", "ds/highlight") if n == ds.hashlabel else " "
				print(template.format(quote(n), name2typ[n], hashdot, prettyminmax(n), format_location(locations.get(n)), c.compression).rstrip())
				if args.location:
					try:
						tf = typed_from(ds, locations[n])
						if tf:
							print(' ' * (13 + len_n + len_t), tf)
					except Exception:
						# source job might be deleted
						pass
			print("    {0} columns".format(fmt_num(len(ds.columns))))
		print("    {0} lines".format(fmt_num(sum(ds.lines))))

		if ds.previous or args.chain:
			chain = ds.chain()
			full_name = 'Full chain' if args.non_empty_chain else 'Chain'
			in_job = len(ds.chain_within_job())
			if in_job > 1:
				if in_job == len(chain):
					in_job = ' (all within job)'
				else:
					in_job = ' ({0} within job)'.format(fmt_num(in_job))
			else:
				in_job = ''
			print("    {0} length {1}{2}, from {3} to {4}".format(full_name, fmt_num(len(chain)), in_job, chain[0], chain[-1]))
			if args.non_empty_chain:
				chain = [ds for ds in chain if sum(ds.lines)]
				print("    Filtered chain length {0}".format(fmt_num(len(chain))))
			if args.chain:
				data = tuple((ix, "%s/%s" % (x.job, x.name), fmt_num(sum(x.lines))) for ix, x in enumerate(chain))
				max_n, max_l = colwidth(x[1:] for x in data)
				template = "{0:3}: {1:%d} ({2:>%d})" % (max_n, max_l)
				printcolwise(data, template, lambda x: (x[0], x[1], x[2]), minrows=8, indent=8)

		if args.slices or args.chainedslices:
			if args.chainedslices and ds.previous:
				data = ((ix, fmt_num(sum(x)), sum(x)) for ix, x in enumerate(zip(*(x.lines for x in ds.chain()))))
				print('    Balance, lines per slice, full chain:')
			else:
				data = ((ix, fmt_num(x), x) for ix, x in enumerate(ds.lines))
				if ds.previous:
					print('    Balance, lines per slice, tip dataset:')
				else:
					print('    Balance, lines per slice:')
			data = sorted(data, key=lambda x: -x[2])
			s = sum(x[2] for x in data)
			len_n = max(len(x[1]) for x in data)
			template = "{0:3}: {1!s}%% ({2:>%d})" % (len_n,)
			printcolwise(data, template, lambda x: (x[0], locale.format_string("%6.2f", (100 * x[2] / (s or 1e20))), x[1]), minrows=8, indent=8)
			print("    Max to average ratio: " + locale.format_string("%2.3f", (max(x[2] for x in data) / ((s or 1e20) / len(data)),) ))

		if ds.previous:
			print("    {0} total lines in chain".format(fmt_num(sum(sum(ds.lines) for ds in chain))))

	finish(badinput)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2024 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from resource import getpagesize
from os import unlink
from os.path import exists
from mmap import mmap
from shutil import copyfileobj
from struct import Struct

from accelerator.compat import unicode, itervalues, PY2

from accelerator.extras import OptionEnum, DotDict, quote
from accelerator.dsutil import typed_writer, typed_reader
from accelerator.error import NoSuchDatasetError
from . import dataset_type

depend_extra = (dataset_type,)

description = r'''
Convert one or more columns in a dataset chain from bytes/ascii/unicode to
any type. Produces one dataset per source dataset if chain_slices=False,
one per slice per source dataset if chain_slices=True (which is slightly
faster to write, slightly slower to read).

Also rehashes if you type the hashlabel, or specify a new hashlabel.
You can also set hashlabel="" to not rehash (and get hashlabel=None).

Without filter_bad the method fails when a value fails to convert and
doesn't have a default. With filter_bad the value is filtered out
together with all other values on the same line.

By default new datasets inherit untouched columns from their parent, but
with filter_bad or when rehashing new independant datasets are produced.
You can set discard_untyped to not copy untouched columns to the new datasets
(and force creating new datasets).

With filter_bad any discarded lines (but only the columns actually typed)
will be saved in separate datasets with ".bad" appended to the name, or
just "bad" for the final one (i.e. not "default.bad").
'''

# Build the +None version of a type name, e.g. number:int => number+None:int
def type_plus_None(typename):
	if ':' in typename:
		a, b = typename.split(':')
		return a + '+None:' + b
	else:
		return typename + '+None'

TYPENAME = OptionEnum(
	list(dataset_type.convfuncs) +
	[type_plus_None(typename) for typename in dataset_type.convfuncs]
)

options = {
	'column2type'               : {'COLNAME': TYPENAME},
	# Anything your system would accept in $TZ. Parses datetimes (but not
	# dates or times) as being in this timezone (you get UTC).
	# No error checking can be done on this (tzset(3) can not return failure).
	# On most 32bit systems this will break dates outside about 1970 - 2037.
	# Setting this will mask most bad dates (mktime(3) "fixes" them).
	# (But "UTC" is special cased and doesn't break anything.)
	# If you set this datetimes will have tzinfo=UTC.
	'timezone'                  : str,
	'hashlabel'                 : str, # leave as None to inherit hashlabel, set to '' to not have a hashlabel
	'defaults'                  : {}, # {'COLNAME': value}, unspecified -> method fails on unconvertible unless filter_bad
	'rename'                    : {}, # {'OLDNAME': 'NEWNAME'} doesn't shadow OLDNAME. (Other COLNAMEs use NEWNAME.)
	                                  # Use {'OLDNAME': None} to discard OLDNAME.
	'caption'                   : 'typed dataset',
	'discard_untyped'           : bool, # Make unconverted columns inaccessible ("new" dataset)
	'filter_bad'                : False, # Discard lines where any column fails typing, saving them in a dataset named "bad"
	'numeric_comma'             : False, # floats as "3,14"
	'length'                    : -1, # Go back at most this many datasets. You almost always want -1 (which goes until previous.source)
	'chain_slices'              : False, # one dataset per slice if rehashing (avoids rewriting at the end)
	'compression'               : 6,     # gzip level
}

datasets = ('source', 'previous',)


byteslike_types = ('bytes', 'ascii', 'unicode',)

cstuff = dataset_type.init()

def prepare(job, slices):
	assert 1 <= options.compression <= 9
	assert datasets.source, 'No source specified'
	if options.timezone:
		cstuff.backend.init(options.timezone.encode('utf-8'))
	else:
		cstuff.backend.init(cstuff.NULL)
	d = datasets.source
	chain = d.chain(stop_ds={datasets.previous: 'source'}, length=options.length)
	res = []
	previous = None
	for ix, ds in enumerate(chain):
		previous = prepare_one(ix, ds, chain, job, slices, previous)
		res.append(previous)
	return res

def prepare_one(ix, source, chain, job, slices, previous_res):
	if ix == len(chain) - 1:
		ds_name = 'default'
	else:
		ds_name = str(ix)
	filename = source.filename
	source_name = source.quoted
	columns = {}
	column2type = dict(options.column2type)
	dup_rename = {}
	just_rename = {}
	rev_rename = {}
	for k, v in options.rename.items():
		if k in source.columns:
			if v in rev_rename:
				raise Exception('Both column %r and column %r rename to %r (in %s)' % (rev_rename[v], k, v, source_name))
			if v is not None:
				rev_rename[v] = k
			if v in column2type:
				dup_rename[k] = v
			else:
				just_rename[k] = v
	renamed_over = set(list(dup_rename.values()) + list(just_rename.values()))
	for k in renamed_over.intersection(dup_rename):
		# renamed over don't need to be duplicated
		just_rename[k] = dup_rename.pop(k)
	if dup_rename:
		dup_ds = source.link_to_here(name='dup.%d' % (ix,), rename=dup_rename, column_filter=dup_rename.values())
		if source.hashlabel in dup_rename:
			just_rename[source.hashlabel] = None
	if just_rename:
		source = source.link_to_here(name='rename.%d' % (ix,), rename=just_rename)
	if dup_rename:
		source = source.merge(dup_ds, name='merge.%d' % (ix,))
	none_support = set()
	for colname, coltype in column2type.items():
		if colname not in source.columns:
			raise Exception("Dataset %s doesn't have a column named %r (has %r)" % (source_name, colname, set(source.columns),))
		dc = source.columns[colname]
		if dc.type not in byteslike_types:
			raise Exception("Dataset %s column %r is type %s, must be one of %r" % (source_name, colname, dc.type, byteslike_types,))
		coltype = coltype.split(':', 1)[0]
		if coltype.endswith('+None'):
			coltype = coltype[:-5]
			none_support.add(colname)
		columns[colname] = dataset_type.typerename.get(coltype, coltype)
		if options.defaults.get(colname, False) is None or dc.none_support:
			none_support.add(colname)
	if options.hashlabel is None:
		hashlabel_override = False
		hashlabel = source.hashlabel
		rehashing = (hashlabel in columns)
	else:
		hashlabel_override = True
		hashlabel = options.hashlabel or None
		rehashing = bool(hashlabel)
	if (options.filter_bad or rehashing) and not options.discard_untyped:
		untyped_columns = set(source.columns)
		untyped_columns -= set(columns) # anything renamed over is irrelevant
		for colname in sorted(untyped_columns):
			dc = source.columns[colname]
			columns[colname] = dc.type
			column2type[colname] = dataset_type.copy_types[dc.type]
			if dc.none_support:
				none_support.add(colname)
	if options.filter_bad or rehashing or options.discard_untyped:
		parent = None
	else:
		parent = source
	if hashlabel and hashlabel not in columns:
		if options.hashlabel:
			raise Exception("Can't rehash %s on discarded column %r." % (source_name, hashlabel,))
		hashlabel = None # it gets inherited from the parent if we're keeping it.
		hashlabel_override = False
	columns = {
		colname: (typ, colname in none_support)
		for colname, typ in columns.items()
	}
	dws = []
	if previous_res:
		# dw or last in dws[] for previous source
		final_previous = previous_res[0] or previous_res[2][-1]
	else:
		final_previous = datasets.previous
	if rehashing:
		previous = final_previous
		for sliceno in range(slices):
			if options.chain_slices and sliceno == slices - 1:
				# This ds will remain and be the tip of the chain for this source
				name = ds_name
			else:
				# This ds is either an earlier part of the chain or will be merged into ds_name in synthesis
				name = '%s.%d' % (ds_name, sliceno,)
			dw = job.datasetwriter(
				columns=columns,
				caption='%s (from %s slice %d)' % (options.caption, source_name, sliceno,),
				hashlabel=hashlabel,
				filename=filename,
				previous=previous,
				meta_only=True,
				name=name,
				for_single_slice=sliceno,
			)
			previous = dw
			dws.append(dw)
	if rehashing and options.chain_slices:
		dw = None
	else:
		dw = job.datasetwriter(
			name=ds_name,
			columns=columns,
			caption='%s (from %s)' % (options.caption, source_name,),
			hashlabel=hashlabel,
			hashlabel_override=hashlabel_override,
			filename=filename,
			parent=parent,
			previous=final_previous,
			meta_only=True,
		)
	if options.filter_bad:
		if previous_res:
			previous = previous_res[1]
		else:
			try:
				previous = datasets.previous.job.dataset('bad')
			except NoSuchDatasetError:
				previous = None
		def best_bad_type(colname):
			dc = source.columns[colname]
			assert dc.type in byteslike_types
			return (dc.type, dc.none_support)
		bad_columns = {name: best_bad_type(name) for name in options.column2type}
		dw_bad = job.datasetwriter(
			name='bad' if ds_name == 'default' else ds_name + '.bad',
			columns=bad_columns,
			previous=previous,
			meta_only=True,
		)
	else:
		dw_bad = None
	return dw, dw_bad, dws, source, source_name, column2type, sorted(columns)


def map_init(vars, name, z='badmap_size'):
	if not vars.badmap_size:
		pagesize = getpagesize()
		line_count = vars.source.lines[vars.sliceno]
		vars.badmap_size = (line_count // 8 // pagesize + 1) * pagesize
		vars.slicemap_size = (line_count * 2 // pagesize + 1) * pagesize
	fh = open(name, 'w+b')
	vars.map_fhs.append(fh)
	fh.truncate(vars[z])
	return fh.fileno()


def analysis(sliceno, slices, prepare_res):
	if options.numeric_comma:
		try_locales = [
			'da_DK', 'nb_NO', 'nn_NO', 'sv_SE', 'fi_FI',
			'en_ZA', 'es_ES', 'es_MX', 'fr_FR', 'ru_RU',
			'de_DE', 'nl_NL', 'it_IT',
		]
		for localename in try_locales:
			localename = localename.encode('ascii')
			if not cstuff.backend.numeric_comma(localename):
				break
			if not cstuff.backend.numeric_comma(localename + b'.UTF-8'):
				break
		else:
			raise Exception("Failed to enable numeric_comma, please install at least one of the following locales: " + " ".join(try_locales))
		dataset_type.numeric_comma = True
	res = [analysis_one(sliceno, slices, p) for p in prepare_res]
	for fn in ('slicemap', 'badmap',):
		fn = '%s%d' % (fn, sliceno,)
		if exists(fn):
			unlink(fn)
	return res

def analysis_one(sliceno, slices, prepare_res):
	dw, dw_bad, dws, source, source_name, column2type, _ = prepare_res
	if source.lines[sliceno] == 0:
		dummy = [0] * slices
		return {}, dummy, {}, {}, dummy
	if dws:
		dw = dws[sliceno]
		rehashing = True
	else:
		rehashing = False
	vars = DotDict(
		sliceno=sliceno,
		slices=slices,
		known_line_count=0,
		badmap_size=0,
		badmap_fd=-1,
		slicemap_size=0,
		slicemap_fd=-1,
		map_fhs=[],
		res_bad_count={},
		res_default_count={},
		res_minmax={},
		first_lap=True,
		rehashing=rehashing,
		hash_lines=None,
		dw=dw,
		dw_bad=dw_bad,
		save_bad=False,
		source=source,
		source_name=source_name,
		column2type=column2type,
	)
	if options.filter_bad:
		vars.badmap_fd = map_init(vars, 'badmap%d' % (sliceno,))
		bad_count, default_count, minmax = analysis_lap(vars)
		if sum(sum(c) for c in itervalues(bad_count)):
			vars.first_lap = False
			vars.res_bad_count = {}
			final_bad_count, default_count, minmax = analysis_lap(vars)
			final_bad_count = [max(c) for c in zip(*final_bad_count.values())]
		else:
			final_bad_count = [0] * slices
	else:
		bad_count, default_count, minmax = analysis_lap(vars)
		final_bad_count = [0] * slices
	for fh in vars.map_fhs:
		fh.close()
	return bad_count, final_bad_count, default_count, minmax, vars.hash_lines


# In python3 indexing into bytes gives integers (b'a'[0] == 97),
# this gives the same behaviour on python2. (For use with mmap.)
class IntegerBytesWrapper(object):
	def __init__(self, inner):
		self.inner = inner
	def close(self):
		self.inner.close()
	def __getitem__(self, key):
		return ord(self.inner[key])
	def __setitem__(self, key, value):
		self.inner[key] = chr(value)

# But even in python3 we can only get int8 support for free,
# and slicemap needs int16.
class Int16BytesWrapper(object):
	_s = Struct('=H')
	def __init__(self, inner):
		self.inner = inner
	def close(self):
		self.inner.close()
	def __getitem__(self, key):
		return self._s.unpack_from(self.inner, key * 2)[0]
	def __setitem__(self, key, value):
		self._s.pack_into(self.inner, key * 2, value)
	def __iter__(self):
		if PY2:
			def it():
				for o in range(len(self.inner) // 2):
					yield self[o]
		else:
			def it():
				for v, in self._s.iter_unpack(self.inner):
					yield v
		return it()


def analysis_lap(vars):
	if vars.rehashing:
		if vars.first_lap:
			out_fn = 'hashtmp.%d' % (vars.sliceno,)
			colname = vars.dw.hashlabel
			coltype = vars.column2type[colname]
			vars.rehashing = False
			real_coltype = one_column(vars, colname, coltype, [out_fn], True)
			vars.rehashing = True
			assert vars.res_bad_count[colname] == [0] # implicitly has a default
			vars.slicemap_fd = map_init(vars, 'slicemap%d' % (vars.sliceno,), 'slicemap_size')
			slicemap = mmap(vars.slicemap_fd, vars.slicemap_size)
			vars.map_fhs.append(slicemap)
			slicemap = Int16BytesWrapper(slicemap)
			hash = typed_writer(real_coltype).hash
			slices = vars.slices
			vars.hash_lines = hash_lines = [0] * slices
			for ix, value in enumerate(typed_reader(real_coltype)(out_fn)):
				dest_slice = hash(value) % slices
				slicemap[ix] = dest_slice
				hash_lines[dest_slice] += 1
			unlink(out_fn)
	for colname, coltype in vars.column2type.items():
		if vars.rehashing:
			out_fns = [vars.dw.column_filename(colname, sliceno=s) for s in range(vars.slices)]
		else:
			out_fns = [vars.dw.column_filename(colname)]
		if options.filter_bad and not vars.first_lap and colname in options.column2type:
			out_fns.append(vars.dw_bad.column_filename(colname, sliceno=vars.sliceno))
			vars.save_bad = True
		else:
			vars.save_bad = False
		one_column(vars, colname, coltype, out_fns)
	return vars.res_bad_count, vars.res_default_count, vars.res_minmax


def one_column(vars, colname, coltype, out_fns, for_hasher=False):
	if for_hasher:
		record_bad = skip_bad = False
	elif vars.first_lap:
		record_bad = options.filter_bad
		skip_bad = False
	else:
		record_bad = 0
		skip_bad = options.filter_bad
	minmax_fn = 'minmax%d' % (vars.sliceno,)

	if coltype.split(':')[0].endswith('+None'):
		coltype = ''.join(coltype.split('+None', 1))
		empty_types_as_None = True
	else:
		empty_types_as_None = False

	fmt = fmt_b = None
	is_null_converter = False
	if coltype in dataset_type.convfuncs:
		shorttype = coltype
		_, cfunc, pyfunc = dataset_type.convfuncs[coltype]
	elif coltype.startswith('null_'):
		shorttype = coltype
		pyfunc = False
		cfunc = True
		is_null_converter = True
	else:
		shorttype, fmt = coltype.split(':', 1)
		_, cfunc, pyfunc = dataset_type.convfuncs[shorttype + ':*']
	if cfunc:
		cfunc = shorttype.replace(':', '_')
	if pyfunc:
		tmp = pyfunc(coltype)
		if callable(tmp):
			pyfunc = tmp
			cfunc = None
		else:
			pyfunc = None
			cfunc, fmt, fmt_b = tmp
	if coltype == 'number':
		cfunc = 'number'
	elif coltype == 'number:int':
		coltype = 'number'
		cfunc = 'number'
		fmt = "int"
	assert cfunc or pyfunc, coltype + " didn't have cfunc or pyfunc"
	coltype = shorttype
	in_fns = []
	in_msgnames = []
	offsets = []
	max_counts = []
	d = vars.source
	assert colname in d.columns, '%s not in %s' % (colname, d.quoted,)
	if not is_null_converter:
		assert d.columns[colname].type in byteslike_types, '%s has bad type in %s' % (colname, d.quoted,)
	in_fns.append(d.column_filename(colname, vars.sliceno))
	in_msgnames.append('%s column %s slice %d' % (d.quoted, quote(colname), vars.sliceno,))
	if d.columns[colname].offsets:
		offsets.append(d.columns[colname].offsets[vars.sliceno])
		max_counts.append(d.lines[vars.sliceno])
	else:
		offsets.append(0)
		max_counts.append(-1)
	if cfunc:
		default_value = options.defaults.get(colname, cstuff.NULL)
		if for_hasher and default_value is cstuff.NULL:
			default_value = None
		default_len = 0
		if default_value is None:
			default_value = cstuff.NULL
			default_value_is_None = True
		else:
			default_value_is_None = False
			if default_value != cstuff.NULL:
				if isinstance(default_value, unicode):
					default_value = default_value.encode("utf-8")
				default_len = len(default_value)
		c = getattr(cstuff.backend, 'convert_column_' + cfunc)
		if vars.rehashing:
			c_slices = vars.slices
		else:
			c_slices = 1
		bad_count = cstuff.mk_uint64(c_slices)
		default_count = cstuff.mk_uint64(c_slices)
		gzip_mode = "wb%d" % (options.compression,)
		if in_fns:
			assert len(out_fns) == c_slices + vars.save_bad
			res = c(*cstuff.bytesargs(in_fns, in_msgnames, len(in_fns), out_fns, gzip_mode, minmax_fn, default_value, default_len, default_value_is_None, empty_types_as_None, fmt, fmt_b, record_bad, skip_bad, vars.badmap_fd, vars.badmap_size, vars.save_bad, c_slices, vars.slicemap_fd, vars.slicemap_size, bad_count, default_count, offsets, max_counts))
			assert not res, 'Failed to convert ' + colname
		vars.res_bad_count[colname] = list(bad_count)
		vars.res_default_count[colname] = sum(default_count)
		coltype = coltype.split(':', 1)[0]
		if is_null_converter:
			dc = d.columns[colname]
			real_coltype = dc.type
			# Some lines may have been filtered out, so these minmax values
			# could be wrong. There's no easy/cheap way to fix that though,
			# and they will never be wrong in the bad direction.
			vars.res_minmax[colname] = [dc.min, dc.max]
		else:
			real_coltype = dataset_type.typerename.get(coltype, coltype)
			if exists(minmax_fn):
				with typed_reader(real_coltype)(minmax_fn) as it:
					vars.res_minmax[colname] = list(it)
				unlink(minmax_fn)
	else:
		# python func
		if for_hasher:
			raise Exception("Can't hash %s on column of type %s." % (vars.source_name, coltype,))
		nodefault = object()
		if colname in options.defaults:
			default_value = options.defaults[colname]
			if default_value is not None:
				default_value = pyfunc(default_value)
		else:
			default_value = nodefault
		if options.filter_bad:
			badmap = mmap(vars.badmap_fd, vars.badmap_size)
			vars.map_fhs.append(badmap)
			if PY2:
				badmap = IntegerBytesWrapper(badmap)
		if vars.rehashing:
			slicemap = mmap(vars.slicemap_fd, vars.slicemap_size)
			vars.map_fhs.append(slicemap)
			slicemap = Int16BytesWrapper(slicemap)
			bad_count = [0] * vars.slices
		else:
			bad_count = [0]
			chosen_slice = 0
		default_count = 0
		dont_minmax_types = {'bytes', 'ascii', 'unicode', 'json', 'complex32', 'complex64'}
		real_coltype = dataset_type.typerename.get(coltype, coltype)
		do_minmax = real_coltype not in dont_minmax_types
		if vars.save_bad:
			bad_fh = typed_writer('bytes')(out_fns.pop(), none_support=True)
		fhs = [typed_writer(real_coltype)(fn, none_support=True) for fn in out_fns]
		if vars.save_bad:
			fhs.append(bad_fh)
		write = fhs[0].write
		col_min = col_max = None
		it = d._column_iterator(vars.sliceno, colname, _type='bytes')
		for ix, v in enumerate(it):
			if vars.rehashing:
				chosen_slice = slicemap[ix]
				write = fhs[chosen_slice].write
			if skip_bad:
				if badmap[ix // 8] & (1 << (ix % 8)):
					bad_count[chosen_slice] += 1
					if vars.save_bad:
						bad_fh.write(v)
					continue
			try:
				if v is None or (empty_types_as_None and v == b''):
					v = None
				else:
					v = pyfunc(v.decode('utf-8'))
			except ValueError:
				if default_value is not nodefault:
					v = default_value
					default_count += 1
				elif record_bad:
					bad_count[chosen_slice] += 1
					bv = badmap[ix // 8]
					badmap[ix // 8] = bv | (1 << (ix % 8))
					continue
				else:
					raise Exception("Invalid value %r with no default in %r in %s" % (v, colname, vars.source_name,))
			if do_minmax and v is not None:
				if col_min is None:
					col_min = col_max = v
				if v < col_min: col_min = v
				if v > col_max: col_max = v
			write(v)
		for fh in fhs:
			fh.close()
		if vars.rehashing:
			slicemap.close()
		if options.filter_bad:
			badmap.close()
		vars.res_bad_count[colname] = bad_count
		vars.res_default_count[colname] = default_count
		vars.res_minmax[colname] = [col_min, col_max]
	return real_coltype

def synthesis(slices, analysis_res, prepare_res):
	# each slice returns [ds0data, ds1data, ...], but we want one list per ds
	analysis_res = zip(*analysis_res)
	for p, a in zip(prepare_res, analysis_res):
		synthesis_one(slices, p, a)

def synthesis_one(slices, prepare_res, analysis_res):
	dw, dw_bad, dws, source, source_name, column2type, columns = prepare_res
	analysis_res = list(analysis_res)
	header_printed = [False]
	def print(msg=''):
		from accelerator.compat import builtins
		if not header_printed[0]:
			if dw:
				ds_name = dw.quoted_ds_name
			else:
				ds_name = quote(dws[0].ds_name[:-1] + '<sliceno>')
			header = '%s -> %s' % (source_name, ds_name)
			builtins.print('%s\n%s' % (header, '=' * len(header)))
			header_printed[0] = True
		builtins.print(msg)
	lines = source.lines
	if options.filter_bad:
		bad_line_count_per_slice = [sum(data[1]) for data in analysis_res]
		lines = [num - b for num, b in zip(lines, bad_line_count_per_slice)]
		bad_line_count_total = sum(bad_line_count_per_slice)
		if bad_line_count_total:
			print()
			print('Bad line count   Column')
			for colname in columns:
				cnt = sum(sum(data[0].get(colname, ())) for data in analysis_res)
				if cnt:
					print('%14d   %s' % (cnt, colname,))
		for s, cnt in enumerate(bad_line_count_per_slice):
			dw_bad.set_lines(s, cnt)
		dw_bad.set_compressions('gzip')
	if options.defaults and sum(sum(data[2].values()) for data in analysis_res):
		print()
		print('Defaulted values')
		for colname in sorted(options.defaults):
			defaulted = [data[2].get(colname, 0) for data in analysis_res]
			if sum(defaulted):
				print('    %s:' % (colname,))
				print('        Slice   Defaulted line count')
				slicecnt = 0
				for sliceno, cnt in enumerate(defaulted):
					if cnt:
						print('        %5d   %d' % (sliceno, cnt,))
						slicecnt += 1
				if slicecnt > 1:
					print('        total   %d' % (sum(defaulted),))
	if dws: # rehashing
		if dw: # not as a chain
			final_bad_count = [data[1] for data in analysis_res]
			hash_lines = [data[4] for data in analysis_res]
			for colname in dw.columns:
				for sliceno in range(slices):
					out_fn = dw.column_filename(colname, sliceno=sliceno)
					with open(out_fn, 'wb') as out_fh:
						for s in range(slices):
							if hash_lines[s][sliceno] - final_bad_count[s][sliceno]:
								src_fn = dws[s].column_filename(colname, sliceno=sliceno)
								with open(src_fn, 'rb') as in_fh:
									copyfileobj(in_fh, out_fh)
			for sliced_dw in dws:
				if sliced_dw:
					sliced_dw.discard()
			for sliceno, counts in enumerate(zip(*[data[4] for data in analysis_res])):
				bad_counts = (data[1][sliceno] for data in analysis_res)
				dw.set_lines(sliceno, sum(counts) - sum(bad_counts))
			for sliceno, data in enumerate(analysis_res):
				dw.set_minmax(sliceno, data[3])
		else:
			for sliceno, data in enumerate(analysis_res):
				if dws[sliceno]:
					dws[sliceno].set_minmax(-1, data[3])
					for s, count in enumerate(data[4]):
						dws[sliceno].set_lines(s, count - data[1][s])
					dws[sliceno].set_compressions('gzip')
	else:
		for sliceno, count in enumerate(lines):
			dw.set_lines(sliceno, count)
		for sliceno, data in enumerate(analysis_res):
			dw.set_minmax(sliceno, data[3])
	if dw:
		dw.set_compressions('gzip')
	if header_printed[0]:
		print()



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2022-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division, print_function

from itertools import cycle
import errno
import os
import sys

from accelerator.colourwrapper import colour
from accelerator.compat import PY2
from accelerator import mp


def split_colour(spec):
	seq, _ = colour.pre_post(spec)
	if seq == '':
		return '', ''
	assert seq.startswith('\x1b[')
	assert seq.endswith('m')
	seq = seq[2:-1]
	assert '\x1b' not in seq
	fg = []
	bg = []
	for part in seq.split(';'):
		code = int(part.split(':', 1)[0])
		if 30 <= code <= 38 or 90 <= code <= 97:
			target = fg
		elif 40 <= code <= 48 or 100 <= code <= 107:
			target = bg
		elif code not in (39, 49):
			print("Sorry, %s can only use colours, not attributes" % (spec,), file=sys.stderr)
			sys.exit(1)
		target.append(part)
	return ';'.join(fg), ';'.join(bg)


# a rather incomplete SGR parser that replaces colour resets by our
# selected colour (if we have one).
def collect_escseq(it, line_fg, line_bg):
	chars = ['\x1b']
	try:
		c = next(it)
		chars.append(c)
		if c == '[':
			while True:
				c = next(it)
				if c == 'm':
					pieces = []
					for piece in ''.join(chars)[2:].split(';'):
						code = int(piece.split(':', 1)[0] or '0', 10)
						if code == 0:
							pieces = ['']
							if line_fg:
								pieces.append(line_fg)
							if line_bg:
								pieces.append(line_bg)
						elif code == 39 and line_fg:
							pieces.append(line_fg)
						elif code == 49 and line_bg:
							pieces.append(line_bg)
						else:
							pieces.append(piece)
					return ('\x1b[', ';'.join(pieces), 'm',)
				chars.append(c)
				if c not in '0123456789;:':
					break
	except (StopIteration, ValueError):
		pass
	return chars


_RC_EPIPE = 124 # process return code used to signal we died with EPIPE

class Liner:
	def __init__(self, process, saved_stdout):
		self.process = process
		self.saved_stdout = saved_stdout

	def close(self):
		os.dup2(self.saved_stdout, 1) # EOF for the liner process (after all children have also exited)
		os.close(self.saved_stdout)
		self.process.join()
		if self.process.exitcode and self.process.exitcode != _RC_EPIPE:
			raise Exception('Liner process exited with %s' % (self.process.exitcode,))


def enable_lines(colour_prefix, lined=True, decode_lines=False, max_count=None, after=0):
	if lined:
		colour._lined = True
		pre_fg0, pre_bg0 = split_colour(colour_prefix + '/oddlines')
		pre_fg1, pre_bg1 = split_colour(colour_prefix + '/evenlines')
		if pre_fg0 == pre_bg0 == pre_fg1 == pre_bg1 == '' and max_count is None:
			return
	else:
		pre_fg0 = pre_bg0 = pre_fg1 = pre_bg1 = ''

	def wrap_EPIPE(*a):
		try:
			return lineme(*a)
		except OSError as e:
			if e.errno == errno.EPIPE:
				exit(_RC_EPIPE)
			raise

	def lineme(lined, max_count, after):
		os.close(liner_w)

		colours = cycle([
			(pre_fg0, pre_bg0),
			(pre_fg1, pre_bg1),
		])

		if PY2:
			in_fh = sys.stdin
			errors = 'replace'
		else:
			in_fh = sys.stdin.buffer.raw
			errors = 'surrogateescape'

		if decode_lines:
			if lined:
				def decode_part(part):
					res = []
					for part in part.split('\\n'):
						part = part.strip('\r')
						if line_bg:
							res.append('\x1b[K')
						res.append(part)
						if line_bg and '\r' not in part:
							res.append('\x1b[K')
						res.append('\n')
					return ''.join(res[:-1]) # final \n is added in the main loop
			else:
				# When not lined the transform should be completely transparent
				def decode_part(part):
					return part.replace('\\n', '\n')

		for line in in_fh:
			line_fg, line_bg = next(colours)
			line = line.strip(b'\r\n').decode('utf-8', errors)
			has_cr = ('\r' in line)
			if max_count is not None:
				if line == '':
					# Empty lines mark the end of output sections, so if we
					# see one when showing the final context we stop.
					if max_count == 0:
						break
					continue
				if line[0] in 'MC': # don't count "I"nfo lines, only "M"atches and "C"ontext
					if line[0] == 'M' and max_count:
						max_count -= 1
					elif max_count == 0 and after > 0:
						after -= 1
				line = line[1:]
			if decode_lines:
				line = '\\'.join(decode_part(part) for part in line.split('\\\\'))
			if lined:
				todo = iter(line)
				data = []
				if line_fg and line_bg:
					data.append('\x1b[%s;%sm' % (line_fg, line_bg,))
				elif line_bg:
					data.append('\x1b[%sm' % (line_bg,))
				elif line_fg:
					data.append('\x1b[%sm' % (line_fg,))
				if line_bg and not decode_lines:
					data.append('\x1b[K') # try to fill the line with bg (if terminal does BCE)
				for c in todo:
					if c == '\x1b':
						data.extend(collect_escseq(todo, line_fg, line_bg))
					else:
						data.append(c)
				if line_bg and not has_cr and not decode_lines:
					# the line might have been long, so if safe and needed try
					# again to fill the line with bg (if terminal does BCE)
					data.append('\x1b[K')
				data.append('\x1b[m\n')
				data = ''.join(data).encode('utf-8', errors)
			else:
				data = line.encode('utf-8', errors) + b'\n'
			while data:
				data = data[os.write(1, data):]
			if max_count is not None and max_count == after == 0:
				break
	liner_r, liner_w = os.pipe()
	liner_process = mp.SimplifiedProcess(
		target=wrap_EPIPE,
		args=(lined, max_count, after,),
		stdin=liner_r,
		name=colour_prefix + '-liner',
	)
	os.close(liner_r)
	saved_stdout = os.dup(1)
	os.dup2(liner_w, 1) # this is stdout for the parent process now
	os.close(liner_w)
	return Liner(liner_process, saved_stdout)



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2023-2024 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test a stand-alone urd for compatibility with old logs and some calls.
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
)

from accelerator.compat import open, url_quote_more
from accelerator.error import UrdPermissionError, UrdConflictError
from accelerator.unixhttp import call
from subprocess import Popen
import json
import os
import time

TEST_LOG_ing = r'''
3|2023-01-01T00:00:00|add|2023-01|test/ing|{}|[["something", "job-0"]]||it's a caption
3|2023-02-01T00:00:01|add|2023-02|test/ing|{"test/ing": {"caption": "it's a caption", "joblist": [["something", "job-0"]], "timestamp": "2023-01"}, "test/two": {"caption": "\"caption\\nwithout escaping\"", "joblist": [["something", "job-1"]], "timestamp": "2023-02"}}|[["something", "job-2"]]||
4	2023-06-19T00:00:00	add	2023-06	test/ing	{}	[["something", "job-3"]]		"caption\nwith lines\nand\ttab\nand \"quotes\" and | too"
4	2024-03-11T00:00:00	add	2024-03	test/ing	{}	[["something", "job-4"]]		"this one has a build job"	"build_job-0"
'''.lstrip()

TEST_LOG_two = r'''
3|2023-02-01T00:00:00|add|2023-02|test/two|{}|[["something", "job-1"]]||"caption\nwithout escaping"
'''.lstrip()

def synthesis(job):
	os.mkdir('urd.db')
	os.mkdir('urd.db/test')
	with open('urd.db/passwd', 'w', encoding='ascii') as fh:
		fh.write('test:pass\n')
	with open('urd.db/test/ing.urd', 'w', encoding='ascii') as fh:
		fh.write(TEST_LOG_ing)
	with open('urd.db/test/two.urd', 'w', encoding='ascii') as fh:
		fh.write(TEST_LOG_two)
	db_path = os.path.realpath('urd.db')
	socket_path = os.path.realpath('socket')
	p = Popen(options.command_prefix + ['urd-server', '--listen', socket_path, '--path', db_path])

	# Allow some time for urd-server to start
	for _ in range(32):
		if os.path.exists(socket_path):
			break
		time.sleep(0.1)

	url = 'unixhttp://' + url_quote_more(socket_path) + '/'
	headers = {'Content-Type': 'application/json', 'Authorization': 'Basic dGVzdDpwYXNz'}
	def check(url_part, want, post_data=None):
		got = call(url + url_part, server_name='urd', data=post_data, headers=headers, fmt=json.loads)
		assert want == got, '\nWanted %r,\ngot    %r' % (want, got,)

	check('list', ['test/ing', 'test/two'])
	check('test/ing/since/0', ['2023-01', '2023-02', '2023-06', '2024-03'])
	check('test/ing/since/2023-01-01', ['2023-02', '2023-06', '2024-03'])
	check('test/ing/since/2023-02', ['2023-06', '2024-03'])
	check('test/ing/first'  , dict(timestamp='2023-01', user='test', build='ing', build_job=None, joblist=[['something', 'job-0']], caption="it's a caption", deps={}))
	check('test/ing/2023-01', dict(timestamp='2023-01', user='test', build='ing', build_job=None, joblist=[['something', 'job-0']], caption="it's a caption", deps={}))
	check('test/two/2023-02', dict(timestamp='2023-02', user='test', build='two', build_job=None, joblist=[['something', 'job-1']], caption=r'"caption\nwithout escaping"', deps={}))
	check('test/ing/2023-02', dict(timestamp='2023-02', user='test', build='ing', build_job=None, joblist=[['something', 'job-2']], caption='', deps={'test/ing': dict(caption="it's a caption", joblist=[['something', 'job-0']], timestamp='2023-01'), 'test/two': dict(caption=r'"caption\nwithout escaping"', joblist=[['something', 'job-1']], timestamp='2023-02')}))
	check('test/ing/2023-06', dict(timestamp='2023-06', user='test', build='ing', build_job=None, joblist=[['something', 'job-3']], caption="caption\nwith lines\nand\ttab\nand \"quotes\" and | too", deps={}))
	check('test/ing/latest', dict(timestamp='2024-03', user='test', build='ing', build_job='build_job-0', joblist=[['something', 'job-4']], caption="this one has a build job", deps={}))
	# Test not overriding, with keys in different order and a build_job.
	check('add', dict(new=False, changed=False, is_ghost=False), b'{"timestamp": "2023-01", "user": "test", "build": "ing", "deps": {}, "joblist": [["something", "job-0"]], "caption": "it\'s a caption", "build_job": "build_job-0"}')
	check('add', dict(new=False, changed=False, is_ghost=False), b'{"user": "test", "build": "ing", "timestamp": "2023-01", "caption": "it\'s a caption", "deps": {}, "joblist": [["something", "job-0"]], "build_job": "build_job-0"}')
	# And not overriding, with only build_job being different.
	check('add', dict(new=False, changed=False, is_ghost=False), b'{"user": "test", "build": "ing", "timestamp": "2024-03", "caption": "this one has a build job", "deps": {}, "joblist": [["something", "job-4"]], "build_job": "changed-0"}')
	# Changing something should not be allowed.
	try:
		check('add', None, b'{"user": "test", "build": "ing", "timestamp": "2023-01", "caption": "new caption", "deps": {}, "joblist": [["something else", "job-0"]], "build_job": "build_job-1"}')
		raise Exception("Changing should not be allowed without the update flag")
	except UrdConflictError:
		pass
	# And now change something with the update flag
	check('add', dict(new=False, changed=True, is_ghost=False, deps=1), b'{"user": "test", "build": "ing", "timestamp": "2023-01", "caption": "new caption", "deps": {}, "joblist": [["something else", "job-0"]], "flags": ["update"], "build_job": "build_job-1"}')
	# And that should ghost the 2023-02 entry that depended on the old version
	check('test/ing/since/0', ['2023-01', '2023-06', '2024-03'])

	# Test truncation
	check('truncate/test/ing/2023-02', dict(count=2, deps=0), '')
	check('test/ing/since/0', ['2023-01'])

	# Try with a user mismatch
	try:
		check('add', dict(new=True, changed=False, is_ghost=False), b'{"user": "wrong", "build": "ing", "timestamp": "2023-06-01", "caption": "", "deps": {}, "joblist": [["something", "job-0"]], "build_job": "build_job-1"}')
		raise Exception("Wrong user was accepted")
	except UrdPermissionError:
		pass
	# And with the wrong password
	headers['Authorization'] = 'Basic dGVzdDpwYXxx'
	try:
		check('add', dict(new=True, changed=False, is_ghost=False), b'{"user": "test", "build": "ing", "timestamp": "2023-06-01", "caption": "", "deps": {}, "joblist": [["something", "job-0"]], "build_job": "build_job-1"}')
		raise Exception("Wrong password was accepted")
	except UrdPermissionError:
		pass

	# And finally verify that we got what we expected in the log.
	with open('urd.db/test/two.urd', 'r', encoding='ascii') as fh:
		assert fh.read() == TEST_LOG_two
	with open('urd.db/test/ing.urd', 'r', encoding='ascii') as fh:
		assert fh.read(len(TEST_LOG_ing)) == TEST_LOG_ing
		want_it = iter([
			'add\t2023-01\ttest/ing\t{}\t[["something else", "job-0"]]\tupdate\t"new caption"\t"build_job-1"\n',
			'truncate\t2023-02\ttest/ing\n',
			'END'
		])
		for got, want in zip(fh, want_it):
			assert got.startswith('4\t'), got
			got = got.split('\t', 2)[2]
			assert want == got, '\nWanted %r,\ngot    %r' % (want, got,)
		assert next(want_it) == 'END'

	p.terminate()
	p.wait()



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2024 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# grep in a dataset(chain)

from __future__ import division, print_function

import sys
import re
import os
from functools import partial
from itertools import chain, repeat, cycle
from collections import deque, defaultdict
from argparse import RawTextHelpFormatter, Action, SUPPRESS, ArgumentError
from multiprocessing import Lock
import json
import datetime
import operator
import signal

from accelerator.compat import unicode, izip, PY2
from accelerator.compat import izip_longest
from accelerator.compat import monotonic
from accelerator.compat import num_types
from accelerator.compat import QueueEmpty
from accelerator.colourwrapper import colour
from .parser import name2ds, ArgumentParser
from accelerator.error import NoSuchWhateverError
from accelerator.extras import DotDict
from accelerator import g
from accelerator import mp


# os.write can't be trusted to write everything
def write(fd, data):
	while data:
		data = data[os.write(fd, data):]


# Some things for --numeric
def number_or_None(obj):
	if isinstance(obj, num_types):
		return obj
	# int first, as float can round
	try:
		return int(obj, 10)
	except ValueError:
		try:
			return float(obj)
		except ValueError:
			# Base 16 has to be handled separately, because using 0 will
			# error on numbers starting with 0 (on python 3).
			# But we have to check for 0x, so things like "a" are not accepted.
			if (isinstance(obj, unicode) and '0x' in obj) or (isinstance(obj, bytes) and b'0x' in obj):
				try:
					return int(obj, 16)
				except ValueError:
					return None

def number_or_error(obj):
	number = number_or_None(obj)
	if number is None:
		raise re.error('%r is not a valid not a number' % (obj,))
	return number

def splitprefix(s, prefixes):
	for p in prefixes:
		if s.startswith(p):
			return p, s[len(p):]
	return None, s

class NumericMatcher:
	# Sufficiently re-compatible interface to number comparisons.
	def __init__(self, pattern):
		import operator
		if pattern == '':
			# Special case that matches any number, but not other things.
			# (.search checks that it's not None, i.e. a number.)
			self.cmp = lambda _: True
		elif ':' in pattern:
			# range, exludes end by default but can include it with :=
			# can also exclude start using <:
			start, stop = pattern.split(':', 1)
			if start.endswith('<'):
				start = number_or_error(start[:-1])
				start = partial(operator.lt, start)
			else:
				start = number_or_error(start)
				start = partial(operator.le, start)
			if stop.startswith('='):
				stop = number_or_error(stop[1:])
				stop = partial(operator.ge, stop)
			else:
				stop = number_or_error(stop)
				stop = partial(operator.gt, stop)
			def in_range(number):
				return start(number) and stop(number)
			self.cmp = in_range
		else:
			prefix, pattern = splitprefix(pattern, ('<=', '<', '>=', '>', '='))
			number = number_or_error(pattern)
			self.cmp = partial({
				None: operator.eq,
				'=' : operator.eq,
				# These are "backwards" because that's how the argument order works out.
				'>' : operator.lt,
				'>=': operator.le,
				'<' : operator.gt,
				'<=': operator.ge,
			}[prefix], number)

	def search(self, number):
		return number is not None and self.cmp(number)

	def finditer(self, s):
		if self.search(number_or_None(s)):
			# Don't include whitespace in the match
			l = s.lstrip()
			r = s.rstrip()
			return (SpanWrap((len(s) - len(l), len(r))),)
		else:
			return ()

class SpanWrap:
	# For .finditer(), to be compatible with re match-objects.
	def __init__(self, span):
		self._span = span
	def span(self):
		return self._span


def main(argv, cfg):
	# -C overrides -A and -B (which in turn override -C)
	class ContextAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			namespace.before_context = namespace.after_context = values

	# Default values for tab_length
	tab_length = DotDict(tab_len=8, field_len=16, min_len=2)

	class TabLengthAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			names = {}
			for suffix in ('_length', 'length', '_len', 'len', ''):
				for name in ('tab', 'field', 'min'):
					names[name + suffix] = name + '_len'
			min_value = dict(tab_len=1, field_len=0, min_len=1)
			order = ['tab_len', 'field_len', 'min_len']
			unnamed = cycle(order)
			values = re.split(r'[/, ]', values)
			if len(values) == 1 and values[0] and '=' not in values[0]:
				# only a single number provided: skip defaults and be traditional
				tab_length.field_len = 0
				tab_length.min_len = 1
			for value in values:
				if not value:
					continue # so you can do -T/ or similar to just activate it
				if '=' in value:
					name, value = value.rsplit('=', 1)
					name = name.lower()
				else:
					name = next(unnamed)
				if name not in names:
					raise ArgumentError(self, 'unknown field %r' % (name,))
				name = names[name]
				try:
					value = int(value)
				except ValueError:
					raise ArgumentError(self, 'invalid int value for %s: %r' % (name, value,))
				if value < min_value[name] or value > 9999:
					raise ArgumentError(self, 'invalid value for %s: %d' % (name, value,))
				tab_length[name] = value
			# -T overrides -t
			namespace.separator = None
			namespace.tab_length = True

	class SeparatorAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			# -t overrides -T
			namespace.tab_length = False
			namespace.separator = values

	parser = ArgumentParser(
		usage="%(prog)s [options] [-e] pattern [...] [-d] ds [...] [[-n] column [...]]",
		description="""positional arguments:
  pattern               (-e, --regexp)
  dataset               (-d, --dataset) can be specified as for "ax ds"
  columns               (-n, --column)""",
		prog=argv.pop(0),
		formatter_class=RawTextHelpFormatter,
	)
	parser.add_argument('-c', '--chain',        action='store_true', negation='dont', help="follow dataset chains", )
	parser.add_argument(      '--chain-length', '--cl', metavar='LENGTH', action='append', type=int, help="follow chains at most this many datasets")
	parser.add_argument(      '--stop-ds',      action='append', metavar='DATASET',   help="follow chains at most to this dataset\nthese options work like in ds.chain() and you can\nspecify them several times (for several datasets)")
	parser.add_argument(      '--colour', '--color', nargs='?', const='always', choices=['auto', 'never', 'always'], type=str.lower, help="colour matched text. can be auto, never or always", metavar='WHEN', )
	parser.add_argument(      '--no-colour', '--no-color', action='store_const', const='never', dest='colour', help=SUPPRESS)
	parser.add_argument(      '--lined',        action='store_true', negation='not',  help="alternate line colour", )
	parser.add_argument('-F', '--fixed-strings',action='store_true', negation='not',  help="patterns are fixed strings, not regular expressions", )
	parser.add_argument('-N', '--numeric',      action='store_true', negation='not',  help="patterns are numbers, or empty to match all numbers\ncan start with a comparison (=, <, <=, >, >=) or be a\nrange (a:b, a<:b, a:=b, a<:=b (incl/excl start/stop))", )
	parser.add_argument('-i', '--ignore-case',  action='store_true', negation='dont', help="case insensitive pattern", )
	parser.add_argument('-v', '--invert-match', action='store_true', negation='dont', help="select non-matching lines", )
	parser.add_argument('-o', '--only-matching',action='store_true', negation='not',  help="only print matching part (or columns with -l)", )
	parser.add_argument('-l', '--list-matching',action='store_true', negation='dont', help="only print matching datasets (or slices with -S)\nwhen used with -o, only print matching columns", )
	parser.add_argument('-m', '--max-count', metavar='NUM', type=int, help="stop after NUM matching lines")
	parser.add_argument('-H', '--headers',      action='store_true', negation='no',   help="print column names before output (and on each change)", )
	parser.add_argument('-O', '--ordered',      action='store_true', negation='not',  help="output in order (one slice at a time)", )
	parser.add_argument('-M', '--allow-missing-columns', action='store_true', negation='dont', help="datasets are allowed to not have (some) columns", )
	parser.add_argument('-g', '--grep',         action='append',                      help="grep this column only, can be specified multiple times", metavar='COLUMN')
	parser.add_argument('+g', '--dont-grep',    action='append',                      help=SUPPRESS)
	parser.add_argument('-r', '--roundrobin',   action='store_true', negation='not',  help="one value at a time across slices")
	parser.add_argument('-s', '--slice',        action='append',                      help="grep this slice only, can be specified multiple times",  type=int)
	parser.add_argument('-u', '--unique',       action='append', nargs='?', metavar='COLUMN', help="only show one value [for this set of columns]")
	parser.add_argument('-D', '--show-dataset', action='store_true', negation='dont', help="show dataset on matching lines", )
	parser.add_argument('-S', '--show-sliceno', action='store_true', negation='dont', help="show sliceno on matching lines", )
	parser.add_argument('-L', '--show-lineno',  action='store_true', negation='dont', help="show lineno (per slice) on matching lines", )
	supported_formats = ('csv', 'raw', 'json',)
	parser.add_argument('-f', '--format', default='csv', choices=supported_formats, help="output format, csv (default) / " + ' / '.join(supported_formats[1:]), metavar='FORMAT', )
	parser.add_argument('-t', '--separator', action=SeparatorAction, help="field separator, default tab / tab-like spaces", )
	parser.add_argument('-T', '--tab-length', metavar='LENGTH', action=TabLengthAction,
		help="field alignment, always uses spaces as separator\n" +
		     "specify as many as you like of TABLEN/FIELDLEN/MINLEN\n" +
		     "or as NAME=VALUE (e.g. \"min=3/field=24\")\n" +
		     "TABLEN works like normal tabs\n" +
		     "FIELDLEN sets a longer minimum between fields\n" +
		     "MINLEN sets a minimum len for all separators\n" +
		     "use \"-T/\" to just activate it (sets %d/%d/%d)" % (tab_length.tab_len, tab_length.field_len, tab_length.min_len,)
	)
	parser.add_argument('-B', '--before-context', type=int, default=0, metavar='NUM', help="print NUM lines of leading context", )
	parser.add_argument('-A', '--after-context',  type=int, default=0, metavar='NUM', help="print NUM lines of trailing context", )
	parser.add_argument('-C', '--context',        type=int, default=0, metavar='NUM', action=ContextAction,
		help="print NUM lines of context\n" +
		     "context is only taken from the same slice of the same\n" +
		     "dataset, and may intermix with output from other\n" +
		     "slices. Use -O to avoid that, or -S -L to see it.",
	)
	parser.add_argument('-e', '--regexp',  default=[], action='append', dest='patterns', help=SUPPRESS)
	parser.add_argument('-d', '--dataset', default=[], action='append', dest='datasets', help=SUPPRESS)
	parser.add_argument('-n', '--column',  default=[], action='append', dest='columns', help=SUPPRESS)
	parser.add_argument('+n', '--not-column',          action='append', dest='not_columns', help=SUPPRESS)
	parser.add_argument('words', nargs='*', help=SUPPRESS)
	args = parser.parse_intermixed_args(argv)

	if args.before_context < 0 or args.after_context < 0:
		print('Context must be >= 0', file=sys.stderr)
		return 1

	if args.fixed_strings and args.numeric:
		# this should be using mutually_exclusive_group, but that doesn't support negations
		print('--numeric not allowed with --fixed-strings', file=sys.stderr)
		return 1

	if args.columns and args.not_columns:
		print("Don't use both --column and --not-column", file=sys.stderr)
		return 1

	if args.grep and args.dont_grep:
		print("Don't use both --grep and --dont-grep", file=sys.stderr)
		return 1

	columns = args.columns
	not_columns = set(args.not_columns or ())

	try:
		args.datasets = [name2ds(cfg, ds) for ds in args.datasets]
	except NoSuchWhateverError as e:
		print(e, file=sys.stderr)
		return 1

	for word in args.words:
		if not args.patterns:
			args.patterns.append(word)
		elif columns and args.datasets:
			columns.append(word)
		else:
			try:
				args.datasets.append(name2ds(cfg, word))
			except NoSuchWhateverError as e:
				if not args.datasets or args.not_columns:
					print(e, file=sys.stderr)
					return 1
				columns.append(word)

	if not args.patterns or not args.datasets:
		parser.print_help(file=sys.stderr)
		return 1

	datasets = args.datasets
	patterns = []
	re_flags = re.DOTALL
	if args.ignore_case:
		re_flags |= re.IGNORECASE
	for pattern in args.patterns:
		if args.fixed_strings:
			pattern = re.escape(pattern)
		try:
			if args.numeric:
				patterns.append(NumericMatcher(pattern))
			else:
				patterns.append(re.compile(pattern, re_flags))
		except re.error as e:
			print("Bad pattern %r:\n%s" % (pattern, e,), file=sys.stderr)
			return 1

	grep_columns = set(args.grep or ())
	dont_grep_columns = set(args.dont_grep or ()) or not_columns
	if grep_columns == set(columns):
		grep_columns = set()

	if args.slice:
		want_slices = []
		for s in args.slice:
			assert 0 <= s < g.slices, "Slice %d not available" % (s,)
			if s not in want_slices:
				want_slices.append(s)
	else:
		want_slices = list(range(g.slices))

	if args.max_count is not None:
		if args.max_count < 1:
			return
		mark_matching_lines = True
	else:
		mark_matching_lines = False

	if len(want_slices) == 1:
		# it will be automatically ordered, so let's not work for it.
		args.ordered = False
		args.roundrobin = False

	if args.list_matching:
		args.roundrobin = False

	if args.roundrobin:
		args.ordered = True

	if args.only_matching:
		if args.list_matching:
			args.list_matching = False
			only_matching = 'columns'
		else:
			only_matching = 'part'
	else:
		only_matching = False

	if args.chain:
		datasets = list(chain.from_iterable(
			ds.chain(length=length, stop_ds=stop_ds)
			for ds, length, stop_ds in zip(
				datasets,
				cycle(args.chain_length or [-1]),
				cycle(args.stop_ds or [None]),
			)
		))

	def columns_for_ds(ds, columns=columns, not_columns=not_columns):
		if columns:
			return [n for n in columns if n in ds.columns]
		else:
			return sorted(n for n in ds.columns if n not in not_columns)

	grep_columns_for_ds = partial(columns_for_ds, columns=grep_columns, not_columns=dont_grep_columns)

	if columns or grep_columns:
		if not args.allow_missing_columns:
			bad = False
			need_cols = set(columns)
			if grep_columns:
				need_cols.update(grep_columns)
			for ds in datasets:
				missing = need_cols - set(ds.columns)
				if missing:
					print('ERROR: %s does not have columns %r' % (ds, missing,), file=sys.stderr)
					bad = True
			if bad:
				return 1
	if columns or not_columns or grep_columns or dont_grep_columns:
		keep_datasets = []
		for ds in datasets:
			if not columns_for_ds(ds):
				continue
			if not grep_columns_for_ds(ds):
				continue
			keep_datasets.append(ds)
		if not keep_datasets:
			return 0
		datasets = keep_datasets

	# For the status reporting, this gives how many lines have been processed
	# when reaching each ds ix, per slice. Ends with an extra fictional ds,
	# i.e. the total number of lines for that slice. And then the same again,
	# to simplify the code in the status shower.
	total_lines_per_slice_at_ds = [[0] * g.slices]
	if args.roundrobin:
		# In the roundrobin case the "sliceno" is the index of the iterating
		# process, not the real sliceno.
		roundrobin_processes = min(g.slices, len(datasets)) # will be set again later to the same value
		status_slices = list(range(roundrobin_processes))
		for ds, fake_sliceno in zip(datasets, cycle(range(roundrobin_processes))):
			total_lines_per_slice_at_ds.append(list(total_lines_per_slice_at_ds[-1]))
			total_lines_per_slice_at_ds[-1][fake_sliceno] += sum(ds.lines[s] for s in want_slices)
	else:
		status_slices = want_slices
		for ds in datasets:
			total_lines_per_slice_at_ds.append([a + b for a, b in zip(total_lines_per_slice_at_ds[-1], ds.lines)])
	total_lines_per_slice_at_ds.append(total_lines_per_slice_at_ds[-1])
	status_interval = {
		# twice per percent, but not too often or too seldom
		sliceno: min(max(total_lines_per_slice_at_ds[-1][sliceno] // 200, 10), 5000)
		for sliceno in status_slices
	}

	# never and always override env settings, auto (default) sets from env/tty
	if args.colour == 'never':
		colour.disable()
		highlight_matches = False
	elif args.colour == 'always':
		colour.enable()
		highlight_matches = True
	else:
		args.colour = 'auto'
		highlight_matches = colour.enabled

	# Don't highlight everything when just trying to cat
	if args.patterns == [''] and not args.numeric:
		highlight_matches = False
	# Don't highlight anything with -l
	if args.list_matching:
		highlight_matches = False
		# Turn off context, so -m works correctly with -l
		args.after_context = args.before_context = 0

	if args.format == 'json':
		# headers was just a mistake, ignore it
		args.headers = False

	separator = args.separator
	if args.tab_length:
		separator = None
	elif separator is None and not sys.stdout.isatty():
		separator = '\t'

	if separator is None:
		# special case where we try to be like a tab, but with spaces.
		# this is useful because terminals typically don't style tabs.
		# and also so you can use smarter tabs and change the length.
		def separate(items, lens, tab_len=tab_length.tab_len, field_len=tab_length.field_len, min_len=tab_length.min_len):
			things = []
			current_pos = min_pos = 0
			for item, item_len in zip(items, lens):
				things.append(item)
				current_pos += item_len
				min_pos += field_len
				spaces = max(min_pos - current_pos, min_len)
				align = (current_pos + spaces) % tab_len
				if align:
					spaces += tab_len - align
				things.append(colour(' ' * spaces, 'grep/separator'))
				current_pos += spaces
			return ''.join(things[:-1])
		separator = '\t'
	else:
		separator_coloured = colour(separator, 'grep/separator')
		def separate(items, lens):
			return separator_coloured.join(items)

	if args.lined or args.max_count:
		from .lined import enable_lines
		liner = enable_lines('grep', lined=args.lined, decode_lines=(args.format == 'raw'), max_count=args.max_count, after=args.after_context)
		if not liner:
			assert not args.max_count
			args.lined = False
	else:
		liner = False

	def json_default(obj):
		if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
			return str(obj)
		elif isinstance(obj, complex):
			return [obj.real, obj.imag]
		else:
			return repr(obj)

	if args.format == 'csv':
		def escape_item(item):
			if item and (separator in item or item[0] in '\'"' or item[-1] in '\'"'):
				return '"' + item.replace('\n', '\\n').replace('\r', '\\r').replace('"', '""') + '"'
			else:
				return item.replace('\n', '\\n').replace('\r', '\\r')
		errors = 'surrogatepass'
	else:
		if args.format == 'raw' and (args.lined or args.max_count):
			# this will be reversed inside the liner process
			def escape_item(item):
				return item.replace('\\', '\\\\').replace('\n', '\\n')
		else:
			escape_item = None
		errors = 'replace' if PY2 else 'surrogateescape'

	if args.unique:
		# A --unique without a value means all, and deletes any previously specified columns.
		while None in args.unique:
			# No .rindex on lists unfortunately.
			args.unique = args.unique[args.unique.index(None) + 1:]
		if args.unique:
			args.unique = set(args.unique)
		else:
			args.unique = True

		# This is a set of tuples, where each value in the tuple is deduped
		# so there is only one copy of it no matter how many tuples it is in.
		# E.g. {('a', 'b', 'c'), ('a', 'b', 'foo')} only has one copy of
		# 'a' and 'b', potentially saving a lot of memory.
		class DedupTupleSet(set):
			__slots__ = ('_dedup',)

			def __init__(self, *a):
				set.__init__(self, *a)
				self._dedup = {}

			def add(self, value):
				value = tuple(self._dedup.setdefault(v, v) for v in value)
				set.add(self, value)

		unique_set = mp.MpSet(_set_cls=DedupTupleSet)
		args.ordered = True
		# This is just a normal object, i.e. not shared between processes.
		# It mirrors everything each process is sure is in the MpSet,
		# to speed up common cases at the cost of higher memory use.
		unique_set_per_process = DedupTupleSet()

	# This is for the ^T handling. Each slice sends an update when finishing
	# a dataset, and every status_interval[sliceno] lines while iterating.
	# To minimise the data sent the only information sent over the queue
	# is (sliceno, finished_dataset).
	# Status printing is triggered by ^T (or SIGINFO if that is available)
	# or by SIGUSR1.
	# Pressing it again within two seconds prints stats per slice too.
	q_status = mp.LockFreeQueue()
	def status_collector():
		q_status.make_reader()
		status = {sliceno: [0, 0] for sliceno in status_slices}
		#            [ds_ix, done_lines]
		total_lines = sum(total_lines_per_slice_at_ds[-1])
		previous = [0]
		# base colour conf in if stderr is a tty, not stdout.
		if args.colour == 'auto':
			colour.configure_from_environ(stdout=sys.stderr)
		def show(sig, frame):
			t = monotonic()
			verbose = (previous[0] + 2 > t) # within 2 seconds of previous
			previous[0] = t
			ds_ixes = []
			progress_lines = []
			progress_fraction = []
			for sliceno in status_slices:
				ds_ix, done_lines = status[sliceno]
				ds_ixes.append(ds_ix)
				max_possible = min(done_lines + status_interval[sliceno], total_lines_per_slice_at_ds[ds_ix + 1][sliceno])
				done_lines = (done_lines + max_possible) / 2 # middle of the possibilities
				progress_lines.append(done_lines)
				total = total_lines_per_slice_at_ds[-1][sliceno]
				if total == 0:
					progress_fraction.append(1)
				else:
					progress_fraction.append(done_lines / total)
			progress_total = sum(progress_lines) / (total_lines or 1)
			bad_cutoff = progress_total - 0.1
			if verbose:
				show_ds = (len(datasets) > 1 and min(ds_ixes) != max(ds_ixes))
				for sliceno, ds_ix, p in zip(status_slices, ds_ixes, progress_fraction):
					if ds_ix == len(datasets):
						msg = 'DONE'
					else:
						msg = '{0:d}% of {1:n} lines'.format(round(p * 100), total_lines_per_slice_at_ds[-1][sliceno])
						if show_ds:
							msg = '%s (in %s)' % (msg, datasets[ds_ix].quoted,)
					msg = '%9d: %s' % (sliceno, msg,)
					if p < bad_cutoff:
						msg = colour(msg, 'grep/infohighlight')
					else:
						msg = colour(msg, 'grep/info')
					write(2, msg.encode('utf-8') + b'\n')
			msg = '{0:d}% of {1:n} lines'.format(round(progress_total * 100), total_lines)
			if len(datasets) > 1:
				min_ds = min(ds_ixes)
				max_ds = max(ds_ixes)
				if min_ds < len(datasets):
					ds_name = datasets[min_ds].quoted
					extra = '' if min_ds == max_ds else ' ++'
					msg = '%s (in %s%s)' % (msg, ds_name, extra,)
			worst = min(progress_fraction)
			if worst < bad_cutoff:
				msg = '%s, worst %d%%' % (msg, round(worst * 100),)
			msg = colour('  SUMMARY: %s' % (msg,), 'grep/info')
			write(2, msg.encode('utf-8') + b'\n')
		for signame in ('SIGINFO', 'SIGUSR1'):
			if hasattr(signal, signame):
				sig = getattr(signal, signame)
				signal.signal(sig, show)
				if hasattr(signal, 'pthread_sigmask'):
					signal.pthread_sigmask(signal.SIG_UNBLOCK, {sig})
		tc_original = None
		using_stdin = False
		if not hasattr(signal, 'SIGINFO') and sys.stdin.isatty():
			# ^T wont work automatically on this OS, so we need to handle it as terminal input
			import termios
			from accelerator.compat import selectors
			sel = selectors.DefaultSelector()
			sel.register(0, selectors.EVENT_READ)
			sel.register(q_status.r, selectors.EVENT_READ)
			try:
				tc_original = termios.tcgetattr(0)
				tc_changed = list(tc_original)
				tc_changed[3] &= ~(termios.ICANON | termios.IEXTEN)
				termios.tcsetattr(0, termios.TCSADRAIN, tc_changed)
				using_stdin = True
			except Exception:
				pass
			# we can't set stdin nonblocking, because it's probably the same
			# file description as stdout, so work around that with alarms.
			def got_alarm(sig, frame):
				raise IOError()
			signal.signal(signal.SIGALRM, got_alarm)
		try:
			while True:
				if using_stdin:
					do_q = False
					for key, _ in sel.select():
						if key.fd == 0:
							try:
								signal.alarm(1) # in case something else read it we block for max 1 second
								try:
									pressed = ord(os.read(0, 1))
								finally:
									signal.alarm(0)
								if pressed == 20:
									write(2, b'\n') # "^T" shows in the terminal
									os.kill(os.getpid(), signal.SIGUSR1)
							except Exception:
								pass
						elif key.fd == q_status.r:
							do_q = True
					if not do_q:
						continue
				try:
					sliceno, finished_dataset = q_status.get()
				except QueueEmpty:
					return
				if finished_dataset:
					ds_ix = status[sliceno][0] + 1
					status[sliceno] = [ds_ix, total_lines_per_slice_at_ds[ds_ix][sliceno]]
				else:
					status[sliceno][1] += status_interval[sliceno]
		finally:
			if tc_original is not None:
				try:
					termios.tcsetattr(0, termios.TCSADRAIN, tc_original)
				except Exception:
					pass
	status_process = mp.SimplifiedProcess(target=status_collector, name='ax grep status')
	children = [status_process]
	# everything else will write, so make it a writer right away
	q_status.make_writer()

	# Output is only allowed while holding this lock, so that long lines
	# do not get intermixed. (Or when alone in producing output.)
	io_lock = Lock()

	# This contains some extra stuff to be a better base for the other
	# outputters.
	# When used directly it enforces no ordering, but merges smaller writes
	# to keep the number of syscalls down.

	class Outputter:
		new_merge_buffer = bytes

		def __init__(self, q_in, q_out):
			self.q_in = q_in
			self.q_out = q_out
			self.buffer = []
			self.merge_buffer = self.new_merge_buffer()

		def put(self, lineno, items, was_match=False):
			data = self.show(lineno, items)
			if mark_matching_lines:
				marker = b'M' if was_match else b'C'
				data = marker + data
			self.merge_buffer += data
			if len(self.merge_buffer) >= 1024:
				self.move_merge()

		def write(self, data):
			write(1, data)

		def move_merge(self):
			if self.merge_buffer:
				with io_lock:
					self.write(self.merge_buffer)
				self.merge_buffer = self.new_merge_buffer()

		def start(self, ds):
			pass

		def end(self, ds):
			if mark_matching_lines:
				self.merge_buffer += b'\n' # empty line == end marker
			self.move_merge()

		def finish(self):
			pass

		def full(self):
			return len(self.buffer) > 5000

		def excite(self):
			self.move_merge()
			if self.buffer:
				self.pump(False)

	# Partially ordered output, each header change acts as a fence.
	# This is used in all slices except the first.
	#
	# The queue gets True when the previous slice is ready for the next
	# header change, and None when the header is printed (and it's ok
	# to resume output).

	class HeaderWaitOutputter(Outputter):
		def start(self, ds):
			expect_ds, show = next(show_headers_here)
			assert ds == expect_ds
			if show:
				self.add_wait()
			else:
				self.excite()

		def add_wait(self):
			# Each sync point is separated by None in the buffer
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

		def move_merge(self):
			data = self.merge_buffer
			self.merge_buffer = self.new_merge_buffer()
			if self.buffer:
				self.pump()
				if self.buffer:
					self.buffer.append(data)
					return
			with io_lock:
				self.write(data)

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				got = self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			if got is True:
				# since pump is only called when we have outputted all
				# currently allowed output or when the next message is an
				# unblock for such output we can just unconditionally send
				# the True on to the next slice here.
				self.q_out.put(True)
				self.pump(wait)
				return
			else:
				self.q_out.put(None)
				self.drain()

		def drain(self):
			assert self.buffer[0] is None, 'The buffer must always stop at a sync point (or empty)'
			with io_lock:
				for pos, data in enumerate(self.buffer[1:], 1):
					if data is None:
						break
					elif data:
						self.write(data)
				else:
					# We did not reach the next fence, so last item is real data
					# and needs to be removed. (The buffer will then be empty and
					# output will continue directly until reaching the sync point.)
					pos += 1
			self.buffer[:pos] = ()

		def finish(self):
			while self.buffer:
				self.pump(True)

	# Partially ordered output, each header change acts as a fence.
	# This is used only in the first slice, and outputs the headers.
	#
	# When it is ready to output headers it sends True in the queue.
	# When the True has travelled around the queue ring all slices are
	# ready, the headers are printed, and None is sent to let the other
	# slices resume output.
	# (When the None returns it is ignored, because output is resumed
	# as soon as the headers are printed.)

	class HeaderOutputter(HeaderWaitOutputter):
		def add_wait(self):
			if not self.buffer:
				self.q_out.put(True)
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain/.put
			self.pump()

		def drain(self):
			assert self.buffer[0] is None, 'The buffer must always stop at a sync point (or empty)'
			with io_lock:
				for pos, data in enumerate(self.buffer[1:], 1):
					if data is None:
						self.q_out.put(True)
						break
					elif data:
						self.write(data)
				else:
					pos += 1
			self.buffer[:pos] = ()

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				got = self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			if got is True:
				# The True we put in when reaching the fence has travelled
				# all the way around the queue ring, it's time to print the
				# new headers
				self.write(next(headers_iter))
				# and then unblock the other slices
				self.q_out.put(None)
				self.drain()
				# No else, when the None comes back we just drop it.
			if not wait:
				self.pump(False)

	# Fully ordered output, each slice waits for the previous slice.
	# For each ds, waits for None (anything really) before starting,
	# sends None when done.

	class OrderedOutputter(Outputter):
		def start(self, ds):
			# Each ds is separated by None in the buffer
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

		def end(self, ds):
			Outputter.end(self, ds)
			if not self.buffer:
				# We are done with this ds, so let next slice continue
				self.q_out.put(None)

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			self.drain()

		def move_merge(self):
			data = self.merge_buffer
			self.merge_buffer = self.new_merge_buffer()
			if self.buffer:
				self.pump()
				if self.buffer:
					self.buffer.append(data)
					return
			# No need for a lock, the other slices aren't writing concurrently.
			self.write(data)

		def drain(self):
			assert self.buffer[0] is None
			for pos, data in enumerate(self.buffer[1:], 1):
				if data is None:
					# We are done with this ds, so let next slice continue
					self.q_out.put(None)
					break
				elif data:
					self.write(data)
			else:
				# We did not reach the next ds, so last item is real data and
				# needs to be removed. (The buffer will then be empty and
				# output will continue directly until reaching the next ds.)
				pos += 1
			self.buffer[:pos] = ()

		def finish(self):
			not_finished = bool(self.buffer)
			while self.buffer:
				self.pump(True)
			if not_finished:
				self.q_out.put(None)

	# Same as above but for the first slice so it prints headers when needed.

	class OrderedHeaderOutputter(OrderedOutputter):
		def start(self, ds):
			# Each ds is separated by None in the buffer
			self.buffer.append(None)
			if headers:
				expect_ds, show = next(show_headers_here)
				assert ds == expect_ds
			else:
				show = False
			if show:
				# Headers changed, start with those.
				self.buffer.append(next(headers_iter))
			else:
				self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

	# This is used as .merge_buffer for unique writers, to keep everything
	# around until it is safe to determine if it's unique or not.

	class UniqueMergeBuffer:
		def __init__(self):
			self.items = []
			self.length = 0

		# So += b'...' still works, for markers
		def __add__(self, data):
			self.append((True, data))
			return self

		def append(self, data):
			self.items.append(data)
			self.length += len(data[1])

		def __len__(self):
			return self.length

		def finish(self):
			for thing, data in self.items:
				if thing is True:
					yield data
				elif thing not in unique_set:
					unique_set.add(thing)
					yield data

	# Choose the right outputter for the kind of sync we need.
	def outputter(q_in, q_out, first_slice=False):
		if args.list_matching:
			cls = Outputter
		elif args.ordered:
			if first_slice:
				cls = OrderedHeaderOutputter
			else:
				cls = OrderedOutputter
			if args.unique:
				# Subclass whichever outputter we needed for unique output.
				class UniqueOutputter(cls):
					new_merge_buffer = UniqueMergeBuffer

					def put(self, lineno, items, was_match=False):
						thing = self.should_output_precheck(items)
						if not thing:
							return
						data = self.show(lineno, items)
						if mark_matching_lines:
							marker = b'M' if was_match else b'C'
							data = marker + data
						self.merge_buffer.append((thing, data))
						if len(self.merge_buffer) >= 1024:
							self.move_merge()

					def write(self, data):
						assert data is not None
						if hasattr(data, 'finish'): # could be plain bytes (headers)
							data = b''.join(data.finish())
						write(1, data)
				cls = UniqueOutputter
		elif headers:
			if first_slice:
				cls = HeaderOutputter
			else:
				cls = HeaderWaitOutputter
		else:
			cls = Outputter
		return cls(q_in, q_out)

	# Make printer for the selected output options
	def make_show(prefix, used_columns):
		def matching_ranges(item):
			ranges = []
			for p in patterns:
				ranges.extend(m.span() for m in p.finditer(item))
			if not ranges:
				return
			# merge overlapping/adjacent ranges
			ranges.sort()
			ranges = iter(ranges)
			start, stop = next(ranges)
			for a, b in ranges:
				if a <= stop:
					stop = max(stop, b)
				else:
					yield start, stop
					start, stop = a, b
			yield start, stop
		def filter_item(item):
			return ''.join(item[a:b] for a, b in matching_ranges(item))
		if args.format == 'json':
			dumps = json.JSONEncoder(ensure_ascii=False, default=json_default).encode
			def show(lineno, items):
				if args.show_sliceno and args.roundrobin:
					(prefix['sliceno'], lineno), items = items
				if only_matching == 'part':
					items = [filter_item(unicode(item)) for item in items]
				if only_matching == 'columns':
					d = {k: v for k, v in zip(used_columns, items) if filter_item(unicode(v))}
				else:
					d = dict(zip(used_columns, items))
				if args.show_lineno:
					prefix['lineno'] = lineno
				if prefix:
					prefix['data'] = d
					d = prefix
				return dumps(d).encode('utf-8', 'surrogatepass') + b'\n'
		else:
			def colour_item(item):
				pos = 0
				parts = []
				for a, b in matching_ranges(item):
					parts.extend((item[pos:a], colour(item[a:b], 'grep/highlight')))
					pos = b
				parts.append(item[pos:])
				return ''.join(parts)
			def show(lineno, items):
				data = list(prefix)
				if args.show_sliceno and args.roundrobin:
					(sliceno, lineno), items = items
					data[-1] = unicode(sliceno)
				if args.show_lineno:
					data.append(unicode(lineno))
				show_items = map(unicode, items)
				if only_matching:
					if only_matching == 'columns':
						show_items = (item if filter_item(item) else '' for item in show_items)
					else:
						show_items = map(filter_item, show_items)
				show_items = list(show_items)
				lens = (len(item) for item in data + show_items)
				if highlight_matches:
					show_items = list(map(colour_item, show_items))
				data.extend(show_items)
				if escape_item:
					lens_unesc = (len(item) for item in data)
					data = list(map(escape_item, data))
					lens_esc = (len(item) for item in data)
					lens = (l + esc - unesc for l, unesc, esc in zip(lens, lens_unesc, lens_esc))
				return separate(data, lens).encode('utf-8', errors) + b'\n'
		return show

	# This is called for each slice in each dataset.
	# Each slice has a separate process (the same for all datasets).
	# The first slice runs in the main process (unless -l), everything
	# else runs from one_slice.

	def grep(ix, ds, sliceno, out):
		out.start(ds)
		if args.roundrobin and ix % roundrobin_processes != sliceno:
			# In roundrobin mode each ds is done fully in a single process.
			out.end(ds)
			return
		if len(patterns) == 1:
			chk = patterns[0].search
		else:
			def chk(s):
				return any(p.search(s) for p in patterns)
		first = [True]
		def mk_iter(col):
			kw = {}
			if first[0]:
				first[0] = False
				if args.roundrobin:
					lines = sum(ds.lines[s] for s in want_slices)
				else:
					lines = ds.lines[sliceno]
				if lines > status_interval[sliceno]:
					def cb(n):
						q_status.put((sliceno, False))
						out.excite()
					kw['callback'] = cb
					kw['callback_interval'] = status_interval[sliceno]
			if ds.columns[col].type == 'ascii':
				kw['_type'] = 'unicode'
			if args.roundrobin:
				todo = [ds._column_iterator(s, col, **kw) for s in want_slices]
				fv = object() # unique marker
				it = (
					v for v in
					chain.from_iterable(izip_longest(*todo, fillvalue=fv))
					if v is not fv
				)
			else:
				it = ds._column_iterator(sliceno, col, **kw)
			if ds.columns[col].type == 'bytes':
				errors = 'replace' if PY2 else 'surrogateescape'
				if ds.columns[col].none_support:
					it = (None if v is None else v.decode('utf-8', errors) for v in it)
				else:
					it = (v.decode('utf-8', errors) for v in it)
			return it
		def mk_slicelineno_iter():
			# Same principle as the --roundrobin handling in mk_iter()
			# but for (sliceno, lineno).
			todo = [
				izip(repeat(sliceno), range(ds.lines[sliceno]))
				for sliceno in want_slices
			]
			fv = object() # unique marker
			return (
				v for v in
				chain.from_iterable(izip_longest(*todo, fillvalue=fv))
				if v is not fv
			)
		used_columns = columns_for_ds(ds)
		used_grep_columns = (grep_columns or dont_grep_columns) and grep_columns_for_ds(ds)
		if used_grep_columns and set(used_grep_columns) != set(used_columns):
			grep_iter = izip(*(mk_iter(col) for col in used_grep_columns))
		else:
			grep_iter = None
		lines_iter = izip(*(mk_iter(col) for col in used_columns))
		if args.show_sliceno and args.roundrobin:
			lines_iter = izip(mk_slicelineno_iter(), lines_iter)
		if args.before_context:
			before = deque((), args.before_context)
		else:
			before = None
		if args.format == 'json':
			prefix = {}
			if args.show_dataset:
				prefix['dataset'] = ds
			if args.show_sliceno:
				prefix['sliceno'] = sliceno
			out.show = make_show(prefix, used_columns)
		else:
			prefix = []
			if args.show_dataset:
				prefix.append(ds)
			if args.show_sliceno:
				prefix.append(str(sliceno))
			prefix = tuple(prefix)
			out.show = make_show(prefix, used_columns)
		if args.invert_match:
			maybe_invert = operator.not_
		else:
			maybe_invert = bool
		if args.numeric:
			fmtfix = number_or_None
		else:
			fmtfix = unicode
		if args.unique:
			if args.unique is True: # all columns
				care_mask = [True] * len(used_columns)
			else: # named columns
				care_mask = [col in args.unique for col in used_columns]
				if not any(care_mask):
					# No columns => no unique items
					out.end(ds)
					return
			if args.ignore_case:
				item_fixup = lambda item: str(item).lower()
			else:
				item_fixup = str
			unique_columns_ix = ds2unique_columns_ix[ds]
			def should_output_precheck(items):
				if args.show_sliceno and args.roundrobin:
					# items is [(sliceno, lineno), actual_items]
					items = items[1]
				items = tuple(
					item_fixup(item)
					for care, item in zip(care_mask, items)
					if care
				)
				thing = (unique_columns_ix,) + items
				if thing in unique_set_per_process:
					return False
				unique_set_per_process.add(thing)
				return thing
			out.should_output_precheck = should_output_precheck
		to_show = 0
		for lineno, items in enumerate(lines_iter):
			if grep_iter:
				grep_items = next(grep_iter)
			elif args.show_sliceno and args.roundrobin:
				grep_items = items[1]
			else:
				grep_items = items
			if maybe_invert(any(chk(fmtfix(item)) for item in grep_items)):
				if q_list:
					try:
						q_list.put((ds, sliceno))
					except mp.QueueClosedError:
						# the main process died, getting a traceback here is not useful
						# (it's probably dead because of -m of | head or something like that)
						pass
					return
				while before:
					out.put(*before.popleft())
				to_show = 1 + args.after_context
				was_match = True
			if to_show:
				out.put(lineno, items, was_match)
				was_match = False
				to_show -= 1
			elif before is not None:
				before.append((lineno, items))
		out.end(ds)

	# This runs in a separate process for each slice except the first
	# one (unless -l), which is handled specially in the main process.

	def one_slice(sliceno, q_in, q_out, q_to_close):
		if q_to_close:
			q_to_close.close()
		if q_in:
			q_in.make_reader()
		if q_out:
			q_out.make_writer()
		if q_list:
			q_list.make_writer()
		try:
			out = outputter(q_in, q_out)
			for ix, ds in enumerate(datasets):
				if seen_list is None or ds not in seen_list:
					grep(ix, ds, sliceno, out)
				q_status.put((sliceno, True))
			out.finish()
		except QueueEmpty:
			# some other process died, no need to print an error here
			# and in fact no need to return an error (either the other
			# process did or we should follow their lead and be happy.)
			sys.exit(0)

	headers_prefix = []
	if args.show_dataset:
		headers_prefix.append('[DATASET]')
	if args.show_sliceno:
		headers_prefix.append('[SLICE]')
	if args.show_lineno:
		headers_prefix.append('[LINE]')

	# [headers] for each ds where headers change (not including the first).
	# this is every ds where sync between slices has to happen when not --ordered.
	# which ds this is is stored in show_headers_here
	headers = []
	if args.headers:
		current_headers = None
		# [(ds, show_headers?)] for each ds in datasets.
		# records ds as well to uncover bugs (it's not really needed)
		show_headers_here = []
		for ds in datasets:
			candidate_headers = columns_for_ds(ds)
			if candidate_headers != current_headers:
				current_headers = candidate_headers
				headers.append(current_headers)
				show_headers_here.append((ds, True))
			else:
				show_headers_here.append((ds, False))
		def gen_headers(headers):
			show_items = headers_prefix + headers
			if escape_item:
				show_items = list(map(escape_item, show_items))
			coloured = (colour(item, 'grep/header') for item in show_items)
			txt = separate(coloured, map(len, show_items))
			if mark_matching_lines:
				txt = 'I' + txt
			return txt.encode('utf-8', 'surrogatepass') + b'\n'
		# remove the starting ds, so no header changes means no special handling.
		current_headers = headers.pop(0)
		if not args.list_matching:
			write(1, gen_headers(current_headers))
		headers_iter = iter(map(gen_headers, headers))
		show_headers_here[0] = (datasets[0], False) # first one is already printed
		show_headers_here = iter(show_headers_here)

	if args.unique:
		# Build a {ds: unique id for the current unique columns}
		# to speed up and save memory in --unique, by replacing the
		# tuple of column names with a unique small integer (shared
		# between processes).
		# Since this is sent to a set in a remote process this should
		# save a decent amount of memory.
		# Also checks that no --unique columns are missing if appropriate.
		unique_columns_ix = 0
		unique_columns2ix = {}
		if args.unique is True:
			unique_filter = lambda _: True
			check_unique_columns_existance = False
		else:
			unique_filter = args.unique.__contains__
			check_unique_columns_existance = not args.allow_missing_columns
		ds2unique_columns_ix = {}
		bad = False
		for ds in datasets:
			unique_columns = tuple(col for col in columns_for_ds(ds) if unique_filter(col))
			if check_unique_columns_existance and len(unique_columns) != len(args.unique):
				missing = args.unique - set(unique_columns)
				print('ERROR: %s does not have columns %r' % (ds.quoted, missing,), file=sys.stderr)
				bad = True
			if unique_columns not in unique_columns2ix:
				unique_columns2ix[unique_columns] = unique_columns_ix
				unique_columns_ix += 1
			ds2unique_columns_ix[ds] = unique_columns2ix[unique_columns]
		if bad:
			return 1

	q_in = q_out = first_q_out = q_to_close = q_list = None
	seen_list = None
	if args.list_matching:
		# in this case all slices get their own process
		# and the main process just prints the maching slices
		q_list = mp.LockFreeQueue()
		separate_process_slices = want_slices
		if not args.show_sliceno:
			seen_list = mp.MpSet()
	else:
		if args.roundrobin:
			separate_process_slices = list(range(1, min(g.slices, len(datasets))))
			roundrobin_processes = len(separate_process_slices) + 1
		else:
			separate_process_slices = want_slices[1:]
		if args.ordered or headers:
			# needs to sync in some way
			q_in = first_q_out = mp.LockFreeQueue()
	for sliceno in separate_process_slices:
		if q_in:
			q_out = mp.LockFreeQueue()
		p = mp.SimplifiedProcess(
			target=one_slice,
			args=(sliceno, q_in, q_out, q_to_close,),
			name='slice-%d' % (sliceno,),
			ignore_EPIPE=bool(liner),
		)
		children.append(p)
		if q_in and q_in is not first_q_out:
			q_in.close()
		q_to_close = first_q_out
		q_in = q_out
	if q_in:
		if q_in is first_q_out:
			# Special case: only one slice, but HeaderOutputter still needs queues.
			class LocalQueue:
				def __init__(self):
					self._lst = []
					self.put = self.put_local = self._lst.append
				def get(self, wait=None):
					if self._lst:
						return self._lst.pop(0)
					raise QueueEmpty()
			q_in = q_out = LocalQueue()
		else:
			q_out = first_q_out
			q_in.make_reader()
			q_out.make_writer()
		if args.ordered:
			q_in.put_local(None)
	del q_to_close
	del first_q_out

	try:
		if args.list_matching:
			if args.headers:
				headers_prefix = ['[DATASET]']
				if seen_list is None:
					headers_prefix.append('[SLICE]')
				write(1, gen_headers([]))
			ordered_res = defaultdict(set)
			q_list.make_reader()
			if seen_list is None:
				used_columns = ['dataset', 'sliceno']
			else:
				used_columns = ['dataset']
			inner_show = make_show({} if args.format == 'json' else [], used_columns)
			def show(ds, sliceno=None):
				if sliceno is None:
					items = [ds]
				else:
					items = [ds, sliceno]
				if mark_matching_lines:
					write(1, b'M')
				write(1, inner_show(None, items))
			while True:
				try:
					ds, sliceno = q_list.get()
				except QueueEmpty:
					break
				if seen_list is None:
					if args.ordered:
						ordered_res[ds].add(sliceno)
					else:
						show(ds, sliceno)
				elif ds not in seen_list:
					seen_list.add(ds)
					if not args.ordered:
						show(ds)
			if args.ordered:
				for ds in datasets:
					if seen_list is None:
						for sliceno in sorted(ordered_res[ds]):
							show(ds, sliceno)
					else:
						if ds in seen_list:
							show(ds)
		else:
			out = outputter(q_in, q_out, first_slice=True)
			sliceno = 0 if args.roundrobin else want_slices[0]
			for ix, ds in enumerate(datasets):
				grep(ix, ds, sliceno, out)
				q_status.put((sliceno, True))
			out.finish()
	except QueueEmpty:
		# don't print an error, probably a subprocess died from EPIPE before
		# the main process. (or the subprocess already printed an error.)
		return 1

	q_status.close()
	if liner:
		try:
			liner.close() # wait for liner to finish before exiting
		except Exception as e:
			print(e, file=sys.stderr)
			return 1
	for c in children:
		c.join()
		if c.exitcode:
			return 1



# -*- coding: utf-8 -*-
############################################################################
#                                                                          #
# Copyright (c) 2024 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from accelerator import subjobs

description = r'''
Test that the expected files are registered (and not registered) when
using job.open(), job.register_file() and job.register_files().

This is the top job, which runs the file producing methods.
'''

def synthesis():
	def test(method, *want_files):
		want_files = set(want_files)
		job = subjobs.build(method)
		got_files = job.files()
		if want_files != got_files:
			extra_files = got_files - want_files
			missing_files = want_files - got_files
			msg = "Got the wrong files from %s: %r" % (method, got_files,)
			if extra_files:
				msg += ", did not expect %r" % (extra_files,)
			if missing_files:
				msg += ", also wanted %r" % (missing_files,)
			raise Exception(msg)

	test('test_register_file_auto', 'analysis slice 0.txt', 'synthesis file.txt', 'result.pickle')
	test('test_register_file_auto_2', 'analysis slice 0.txt', 'synthesis file.txt')
	test('test_register_file_manual', 'analysis slice 0.txt', 'analysis slice 2.txt', 'registered temp file.txt', 'synthesis file.txt', 'subdir/deep/file.txt')
